from collections import Counter, namedtuple
import copy
import itertools as it
import re
import shlex
import subprocess
import tempfile
import sys
import torch

import h5py
import numpy as np
from scipy.spatial.distance import pdist
from scipy.spatial.transform import Rotation
import rmsd

try:
    from thermoanalysis.QCData import QCData
    from thermoanalysis.thermo import thermochemistry
except ModuleNotFoundError:
    pass

from pysisyphus import logger
from pysisyphus.config import p_DEFAULT, T_DEFAULT
from pysisyphus.constants import BOHR2ANG
from pysisyphus.elem_data import (
    MASS_DICT,
    ISOTOPE_DICT,
    ATOMIC_NUMBERS,
    COVALENT_RADII as CR,
    VDW_RADII as VDWR,
)
from pysisyphus.helpers_pure import (
    eigval_to_wavenumber,
    full_expand,
    molecular_volume,
    to_subscript_num,
)
from pysisyphus.intcoords import (
    DLC,
    HDLC,
    RedundantCoords,
    TRIC,
    TMTRIC,
    HybridRedundantCoords,
    CartesianCoords,
    MWCartesianCoords,
)
from pysisyphus.intcoords.exceptions import (
    NeedNewInternalsException,
    RebuiltInternalsException,
    DifferentCoordLengthsException,
)
from pysisyphus.intcoords.helpers import get_tangent
from pysisyphus.intcoords.setup import BOND_FACTOR
from pysisyphus.intcoords.setup_fast import find_bonds
from pysisyphus.xyzloader import make_xyz_str


def inertia_tensor(coords3d, masses):
    """Inertita tensor.

                          | x² xy xz |
    (x y z)^T . (x y z) = | xy y² yz |
                          | xz yz z² |
    """
    x, y, z = coords3d.T
    squares = np.sum(coords3d**2 * masses[:, None], axis=0)
    I_xx = squares[1] + squares[2]
    I_yy = squares[0] + squares[2]
    I_zz = squares[0] + squares[1]
    I_xy = -np.sum(masses * x * y)
    I_xz = -np.sum(masses * x * z)
    I_yz = -np.sum(masses * y * z)
    I = np.array(((I_xx, I_xy, I_xz), (I_xy, I_yy, I_yz), (I_xz, I_yz, I_zz)))
    return I


def get_trans_rot_vectors(cart_coords, masses, rot_thresh=1e-6):
    """Vectors describing translation and rotation.

    These vectors are used for the Eckart projection by constructing
    a projector from them.

    See Martin J. Field - A Pratcial Introduction to the simulation
    of Molecular Systems, 2007, Cambridge University Press, Eq. (8.23),
    (8.24) and (8.26) for the actual projection.

    See also https://chemistry.stackexchange.com/a/74923.

    Parameters
    ----------
    cart_coords : np.array, 1d, shape (3 * atoms.size, )
        Atomic masses in amu.
    masses : iterable, 1d, shape (atoms.size, )
        Atomic masses in amu.

    Returns
    -------
    ortho_vecs : np.array(6, 3*atoms.size)
        2d array containing row vectors describing translations
        and rotations.
    """

    coords3d = np.reshape(cart_coords, (-1, 3))
    total_mass = masses.sum()
    com = 1 / total_mass * np.sum(coords3d * masses[:, None], axis=0)
    coords3d_centered = coords3d - com[None, :]

    I = inertia_tensor(coords3d, masses)
    _, Iv = np.linalg.eigh(I)
    Iv = Iv.T

    masses_rep = np.repeat(masses, 3)
    sqrt_masses = np.sqrt(masses_rep)
    num = len(masses)

    def get_trans_vecs():
        """Mass-weighted unit vectors of the three cartesian axes."""

        for vec in ((1, 0, 0), (0, 1, 0), (0, 0, 1)):
            _ = sqrt_masses * np.tile(vec, num)
            yield _ / np.linalg.norm(_)

    def get_rot_vecs():
        """As done in geomeTRIC."""

        rot_vecs = np.zeros((3, cart_coords.size))
        # p_vecs = Iv.dot(coords3d_centered.T).T
        for i in range(masses.size):
            p_vec = Iv.dot(coords3d_centered[i])
            for ix in range(3):
                rot_vecs[0, 3 * i + ix] = Iv[2, ix] * p_vec[1] - Iv[1, ix] * p_vec[2]
                rot_vecs[1, 3 * i + ix] = Iv[2, ix] * p_vec[0] - Iv[0, ix] * p_vec[2]
                rot_vecs[2, 3 * i + ix] = Iv[0, ix] * p_vec[1] - Iv[1, ix] * p_vec[0]
        rot_vecs *= sqrt_masses[None, :]
        return rot_vecs

    trans_vecs = list(get_trans_vecs())
    rot_vecs = np.array(get_rot_vecs())
    # Drop vectors with vanishing norms
    rot_vecs = rot_vecs[np.linalg.norm(rot_vecs, axis=1) > rot_thresh]
    tr_vecs = np.concatenate((trans_vecs, rot_vecs), axis=0)
    tr_vecs = np.linalg.qr(tr_vecs.T)[0].T
    return tr_vecs


def get_trans_rot_projector(cart_coords, masses, full=False):
    tr_vecs = get_trans_rot_vectors(cart_coords, masses=masses)
    U, s, _ = np.linalg.svd(tr_vecs.T)
    if full:
        P = np.eye(cart_coords.size)
        for tr_vec in tr_vecs:
            P -= np.outer(tr_vec, tr_vec)
    else:
        P = U[:, s.size :].T
    return P


class Geometry:
    coord_types = {
        "cart": None,
        "redund": RedundantCoords,
        "hredund": HybridRedundantCoords,
        "dlc": DLC,
        "hdlc": HDLC,
        "tric": TRIC,
        "tmtric": TMTRIC,
        "cartesian": CartesianCoords,
        "mwcartesian": MWCartesianCoords,
    }

    def __init__(
        self,
        atoms,
        coords,
        fragments=None,
        coord_type="cart",
        coord_kwargs=None,
        isotopes=None,
        freeze_atoms=None,
        comment="",
        name="",
    ):
        """Object representing atoms in a coordinate system.

        The Geometry represents atoms and their positions in coordinate
        system. By default cartesian coordinates are used, but internal
        coordinates are also possible.

        Parameters
        ----------
        atoms : iterable
            Iterable of length N, containing  element symbols.
        coords : 1d iterable
            1d iterable of length 3N, containing the cartesian coordinates
            of N atoms.
        fragments : dict, optional
            Dict with different keys denoting different fragments. The values
            contain lists of atom indices.
        coord_type : {"cart", "redund"}, optional
            Type of coordinate system to use. Right now cartesian (cart)
            and redundand (redund) are supported.
        coord_kwargs : dict, optional
            Dictionary containing additional arguments that get passed
            to the constructor of the internal coordinate class.
        isotopes : iterable of pairs, optional
            Iterable of pairs consisting of 0-based atom index and either an integer
            or a float. If an integer is given the closest isotope mass will be selected.
            Given a float, this float will be directly used as mass.
        freeze_atoms : iterable of integers
            Specifies which atoms should remain fixed at their initial positions.
        comment : str, optional
            Comment string.
        name : str, optional
            Verbose name of the geometry, e.g. methanal or water. Used for printing
        """
        self.atoms = tuple([atom.capitalize() for atom in atoms])
        # self._coords always holds cartesian coordinates.
        self._coords = np.array(coords, dtype=float).flatten()
        assert self._coords.size == (3 * len(self.atoms)), (
            f"Expected 3N={3*len(self.atoms)} cartesian coordinates but got "
            f"{self._coords.size}. Did you accidentally supply internal "
            "coordinates?"
        )
        if fragments is None:
            fragments = dict()
        self.fragments = fragments
        self.coord_type = coord_type
        if coord_kwargs is None:
            coord_kwargs = dict()
        self.coord_kwargs = coord_kwargs
        if isotopes is None:
            isotopes = list()
        self.isotopes = isotopes
        if freeze_atoms is None:
            freeze_atoms = list()
        elif type(freeze_atoms) is str:
            freeze_atoms = full_expand(freeze_atoms)
        self.freeze_atoms = np.array(freeze_atoms, dtype=int)
        self.comment = comment
        self.name = name

        self._masses = None
        self._energy = None
        self._forces = None
        self._hessian = None
        self.within_partial_hessian = None
        self._all_energies = None
        self.calculator = None

        assert (
            # Negative atom indices are not allowed.
            all(self.freeze_atoms >= 0)
            and (
                # Allow an empty array, no frozen atoms.
                (self.freeze_atoms.size == 0)
                # Or check that the biggest index is still in the valid range
                or (self.freeze_atoms.max() < len(self.atoms))
            )
        ), f"'freeze_atoms' must all be >= 0 and < {len(self.atoms)}!"

        # Disallow any coord_kwargs with coord_type == 'cart'
        if (coord_type == "cart") and not (coord_kwargs is None or coord_kwargs == {}):
            print(
                "coord_type is set to 'cart' but coord_kwargs were given. "
                "This is probably not intended. Exiting!"
            )
            sys.exit()

        # Coordinate systems are handled below
        coord_class = self.coord_types[self.coord_type]
        if coord_class:
            if (len(self.freeze_atoms) > 0) and ("freeze_atoms" not in coord_kwargs):
                coord_kwargs["freeze_atoms"] = freeze_atoms
            self.internal = coord_class(
                atoms,
                self.coords3d.copy(),
                masses=self.masses,
                **coord_kwargs,
            )
        else:
            self.internal = None

    @property
    def moving_atoms(self):
        return [atom for i, atom in enumerate(self.atoms) if i not in self.freeze_atoms]

    def moving_atoms_jmol(self):
        atoms = list()
        freeze_atoms = self.freeze_atoms
        for i, atom in enumerate(self.atoms):
            atom = atom if i not in freeze_atoms else "X"
            atoms.append(atom)
        self.jmol(atoms=atoms)

    @property
    def sum_formula(self):
        unique_atoms = sorted(set(self.atoms))
        counter = Counter(self.atoms)
        atoms = list()
        num_strs = list()

        def set_atom(atom):
            atoms.append(atom)
            num = counter[atom]
            if num == 1:
                num_str = ""
            else:
                num_str = to_subscript_num(num)
            num_strs.append(num_str)

        # Hill-System
        for atom in ("C", "H"):
            try:
                unique_atoms.remove(atom)
                set_atom(atom)
            except ValueError:
                pass
        for atom in unique_atoms:
            set_atom(atom)

        return "".join([f"{atom}{num_str}" for atom, num_str in zip(atoms, num_strs)])

    def assert_compatibility(self, other):
        """Assert that two Geometries can be substracted from each other.

        Parameters
        ----------
        other : Geometry
            Geometry for comparison.
        """
        same_atoms = self.atoms == other.atoms
        same_coord_type = self.coord_type == other.coord_type
        same_coord_length = len(self.coords) == len(other.coords)
        assert same_atoms, "Atom number/ordering is incompatible!"
        assert same_coord_type, "coord_types are incompatible!"
        try:
            assert same_coord_length, "Different length of coordinate vectors!"
        except AssertionError:
            raise DifferentCoordLengthsException

    def __eq__(self, other):
        return (self.atoms == other.atoms) and np.allclose(
            self.coords, other.coords, atol=1e-8
        )

    def __sub__(self, other):
        self.assert_compatibility(other)
        if self.coord_type in ("cart", "cartesian"):
            diff = self.coords - other.coords
        elif self.coord_type in ("redund", "dlc"):
            # Take periodicity of dihedrals into account by calling
            # get_tangent(). Care has to be taken regarding the orientation
            # of the returned tangent vector. It points from self to other.
            #
            # As we want to return the difference between two vectors we
            # have to reverse the direction of the tangent by multiplying it
            # with -1 to be consistent with basic subtraction laws ...
            # A - B = C, where C is a vector pointing from B to A (B + C = A)
            # In our case get_tangent returns B - A, that is a vector pointing
            # from A to B.
            diff = -get_tangent(
                self.internal.prim_coords,
                other.internal.prim_coords,
                self.internal.dihedral_indices,
            )
        else:
            raise Exception("Invalid coord_type!")

        # Convert to DLC
        if self.coord_type == "dlc":
            diff = self.internal.U.T.dot(diff)
        return diff

    def __add__(self, other):
        atoms = tuple(self.atoms) + tuple(other.atoms)
        coords = np.concatenate((self.cart_coords, other.cart_coords))
        return Geometry(atoms, coords)

    def atom_xyz_iter(self):
        return iter(zip(self.atoms, self.coords3d))

    def copy(self, coord_type=None, coord_kwargs=None):
        """Returns a new Geometry object with same atoms and coordinates.

        Parameters
        ----------
        coord_type : str
            Desired coord_type, defaults to current coord_type.

        coord_kwargs : dict, optional
            Any desired coord_kwargs that will be passed to the RedundantCoords
            object.
        Returns
        -------
        geom : Geometry
            New Geometry object with the same atoms and coordinates.
        """
        if coord_type is None:
            coord_type = self.coord_type

        if coord_kwargs is None:
            coord_kwargs = dict()

        # Geometry constructor will exit when coord_kwargs are given
        # with coord_type == 'cart'. So we only supply it when we are
        # NOT using cartesian coordinates.
        _coord_kwargs = None
        if coord_type != "cart":
            try:
                typed_prims = self.internal.typed_prims
            # Will be raised if the current coord_type is 'cart'
            except AttributeError:
                typed_prims = None
            _coord_kwargs = {
                "typed_prims": typed_prims,
                "check_bends": True,
            }
            _coord_kwargs.update(coord_kwargs)
        return Geometry(
            self.atoms,
            self._coords.copy(),
            coord_type=coord_type,
            coord_kwargs=_coord_kwargs,
            isotopes=copy.deepcopy(self.isotopes),
            freeze_atoms=self.freeze_atoms.copy(),
        )

    def copy_all(self, coord_type=None, coord_kwargs=None):
        new_geom = self.copy(coord_type, coord_kwargs)
        new_geom.set_calculator(self.calculator)
        new_geom.energy = self._energy
        if self._forces is not None:
            new_geom.cart_forces = self._forces
        if self._hessian is not None:
            new_geom.cart_hessian = self._hessian
        return new_geom

    def atom_indices(self):
        """Dict with atom types as key and corresponding indices as values.

        Returns
        -------
        inds_dict : dict
            Unique atom types as keys, corresponding indices as values.
        """
        inds_dict = {}
        for atom_type in set(self.atoms):
            inds_dict[atom_type] = [
                i for i, atom in enumerate(self.atoms) if atom == atom_type
            ]
        return inds_dict

    @property
    def atom_types(self):
        return set(self.atoms)

    @property
    def atomic_numbers(self):
        return [ATOMIC_NUMBERS[a.lower()] for a in self.atoms]

    def get_fragments(self, regex):
        regex = re.compile(regex)
        frags = [frag for frag in self.fragments.keys() if regex.search(frag)]
        org_indices = list(it.chain(*[self.fragments[frag] for frag in frags]))

        new_atoms = [self.atoms[ind] for ind in org_indices]
        new_coords = self.coords3d[org_indices].copy()
        new_fragments = dict()
        i = 0
        for frag in frags:
            frag_atoms = len(self.fragments[frag])
            new_fragments[frag] = list(range(i, i + frag_atoms))
            i += frag_atoms
        return Geometry(new_atoms, new_coords, fragments=new_fragments)

    @property
    def layers(self):
        try:
            layers = self.calculator.layers
        except AttributeError:
            layers = (None,)
        return layers

    def del_atoms(self, inds, **kwargs):
        atoms = [atom for i, atom in enumerate(self.atoms) if not (i in inds)]
        c3d = self.coords3d
        coords3d = np.array(
            [c3d[i] for i, _ in enumerate(self.atoms) if not (i in inds)]
        )
        return Geometry(atoms, coords3d.flatten(), **kwargs)

    def set_calculator(self, calculator, clear=True):
        """Reset the object and set a calculator."""
        if clear:
            self.clear()
        self.calculator = calculator

        if hasattr(self.calculator, "freeze_atoms"):
            self.calculator.freeze_atoms = self.freeze_atoms.copy()

    @property
    def is_analytical_2d(self):
        try:
            return self.calculator.analytical_2d
        except AttributeError:
            return False

    @property
    def mm_inv(self):
        """Inverted mass matrix.

        Returns a diagonal matrix containing the inverted atomic
        masses.
        """
        return np.diag(1 / self.masses_rep)

    @property
    def mm_sqrt_inv(self):
        """Inverted square root of the mass matrix."""
        return np.diag(1 / (self.masses_rep**0.5))

    @property
    def coords(self):
        """1d vector of atomic coordinates.

        Returns
        -------
        coords : np.array
            1d array holding the current coordinates.
        """
        if self.internal:
            coords = self.internal.coords
        else:
            # self._coords will always hold Cartesian coordinates.
            coords = self._coords
        return coords

    def set_coord(self, ind, coord):
        """Set a coordinate by index.

        Parameters
        ----------
        ind : int
            Index in of the coordinate to set in the self.coords array.
        coord : float
            Coordinate value.
        """
        assert (
            self.coord_type == "cart" and len(self.freeze_atoms) == 0
        ), "set_coord was not yet tested with coord_type != 'cart' and frozen atoms!"
        self.coords[ind] = coord
        self.clear()

    def set_coords(self, coords, cartesian=False, update_constraints=False):
        coords = np.array(coords).flatten()

        # Do Internal->Cartesian backtransformation if internal coordinates are used.
        if self.internal:
            # When internal coordinates are employed it may happen, that the underlying
            # Cartesian coordinates are updated, e.g. from the IPIServer calculator, which
            # may yield different internal coordinates.
            #
            # Here we update the Cartesians of the internal coordinate object to the new
            # values and calculate new internal coordinates, from which we can derive a step
            # in internals.
            if cartesian:
                self.assert_cart_coords(coords)
                cart_coords = coords.copy()
                # Update Cartesians of internal coordinate object and calculate
                # new internals.
                self.internal.coords3d = coords
                # Determine new internal coordinates, so we can later calculate a
                # step in internal coordinates.
                coords = self.internal.coords
                # Finally we also update the Cartesian coordinates of the Geometry object,
                # so the subsequent sanity check does not fail. This also allows updating
                # the coordiantes of atoms that are frozen. We set Geometry._coords directly,
                # instead of Geometry.cart_coords or Geometry.coords3d, to avoid an infinite
                # recursion.
                self._coords = cart_coords

            # Sanity check, asserting that the cartesian coordinates of the
            # Geometry object and the internal coordinate object are the same.
            np.testing.assert_allclose(self.coords3d, self.internal.coords3d)

            try:
                int_step = coords - self.internal.coords
                cart_step = self.internal.transform_int_step(
                    int_step, update_constraints=update_constraints
                )
                # From now on coords will always hold Cartesian coordinates!
                coords = self._coords + cart_step
            except NeedNewInternalsException as exception:
                invalid_inds = exception.invalid_inds
                # Check if the remaining internal coordinates are valid
                valid_typed_prims = [
                    typed_prim
                    for i, typed_prim in enumerate(self.internal.typed_prims)
                    if i not in invalid_inds
                ]
                coords3d = exception.coords3d.copy()
                coord_class = self.coord_types[self.coord_type]
                coord_kwargs = self.coord_kwargs.copy()
                """Instead of using only the remaining, valid typed_prims
                we could look for an entirely new set of typed_prims.
                
                But when we do this and we end up with more coordinates
                than before, this will lead to problems with the HDF5 dump.
                
                No problems arise when fewer coordinates are used
                (valid_typed_prims <= self.internal.typed_prims).
                With typed prims, only the remaining, valid typed_prims
                will be defined for the new geometry.
                
                coord_kwargs["typed_prims"] = valid_typed_prims # Currently disabled
                
                With 'define_prims' the remaining, valid typed_prims
                will be used, together with newly determined internal
                coordinates. This supports, e.g., the switch from a simple
                bend to a linear bend and its complement.
                
                Currently the default."""
                coord_kwargs["define_prims"] = valid_typed_prims

                self.internal = coord_class(self.atoms, coords3d, **coord_kwargs)
                self._coords = coords3d.flatten()
                raise RebuiltInternalsException(
                    typed_prims=self.internal.typed_prims.copy()
                )

        # Restore original coordinates of frozen atoms. Right now this should
        # be redundant, as the Cartesian step is also constrainted in the
        # Internal->Cartesian backtransformation. But we keep it for now.
        coords.reshape(-1, 3)[self.freeze_atoms] = self.coords3d[self.freeze_atoms]
        # Set new Cartesian coordinates
        self._coords = coords
        # Reset all values because no calculations with the new coords
        # have been performed yet.
        self.clear()

    def reset_coords(self, new_typed_prims=None):
        if self.coord_type == "cart":
            return

        coord_class = self.coord_types[self.coord_type]
        self.internal = coord_class(
            self.atoms, self.coords3d, typed_prims=new_typed_prims
        )

    @coords.setter
    def coords(self, coords):
        """Wrapper for saving coordinates internally.

        Parameters
        ----------
        coords : np.array
            1d array containing atomic coordiantes. It's length
            depends on the coordinate system.
        """
        self.set_coords(coords)

    @property
    def coords3d(self):
        """Coordinates in 3d.

        Returns
        -------
        coords3d : np.array
            Coordinates of the Geometry as 2D array.
        """
        return self._coords.reshape(-1, 3)

    @coords3d.setter
    def coords3d(self, coords3d):
        self.set_coords(coords3d, cartesian=True)

    @property
    def cart_coords(self):
        return self._coords

    @cart_coords.setter
    def cart_coords(self, coords):
        self.set_coords(coords, cartesian=True)

    @property
    def coords_by_type(self):
        """Coordinates in 3d by atom type and their corresponding indices.

        Returns
        -------
        cbt : dict
            Dictionary with the unique atom types of the Geometry as keys.
            It's values are the 3d coordinates of the corresponding atom type.
        inds : dict
            Dictionary with the unique atom types of the Geometry as keys.
            It's values are the original indices of the 3d coordinates in the
            whole coords3d array.
        """
        cbt = dict()
        inds = dict()
        # for i, (atom, c3d) in enumerate(zip(self.atoms, self.coords3d)):
        # cbt.setdefault(atom, list()).append((i, c3d.tolist()))
        for i, (atom, c3d) in enumerate(zip(self.atoms, self.coords3d)):
            cbt.setdefault(atom, list()).append((c3d))
            inds.setdefault(atom, list()).append(i)
        for atom, c3d in cbt.items():
            cbt[atom] = np.array(c3d)
            inds[atom] = np.array(inds[atom])
        return cbt, inds

    @property
    def comment(self):
        en_width = 20
        # Check if we have to drop an (old) energy entry
        try:
            _ = float(self._comment[:en_width])
            # Drop old energy entry
            self._comment = self._comment[en_width + 2 :]
        except (ValueError, IndexError):
            pass

        # Prepend (new) energy, if present
        if self._energy:
            en_str = f"{self._energy: >{en_width}.8f} , "
        else:
            en_str = ""
        return f"{en_str}{self._comment}"

    @comment.setter
    def comment(self, new_comment):
        self._comment = new_comment

    @property
    def masses(self):
        if self._masses is None:
            # Lookup tabuled masses in internal database
            masses = np.array([MASS_DICT[atom.lower()] for atom in self.atoms])
            # Use (different) isotope masses if requested
            for atom_index, iso_mass in self.isotopes:
                if "." not in str(iso_mass):
                    atom = self.atoms[atom_index].lower()
                    key = (atom, iso_mass)
                    try:
                        iso_mass = ISOTOPE_DICT[key]
                    except KeyError as err:
                        print(
                            f"Found no suitable mass for '{atom.capitalize()}' with approx. "
                            f"mass of ~{iso_mass} au!"
                        )
                        raise err
                masses[atom_index] = float(iso_mass)
            self.masses = masses
        return self._masses

    @masses.setter
    def masses(self, masses):
        assert len(masses) == len(self.atoms)
        masses = np.array(masses, dtype=float)
        self._masses = masses
        # Also try to propagate updated masses to the internal coordiante object
        try:
            self.internal.masses = masses
        except AttributeError:
            pass

    @property
    def masses_rep(self):
        # Some of the analytical potentials are only 2D
        repeat_masses = 2 if (self._coords.size == 2) else 3
        return np.repeat(self.masses, repeat_masses)

    @property
    def total_mass(self):
        return sum(self.masses)

    def center_of_mass_at(self, coords3d):
        """Returns the center of mass at given coords3d.

        Parameters
        ----------
        coords3d : np.array, shape(N, 3)
            Cartesian coordiantes.

        Returns
        -------
        R : np.array, shape(3, )
            Center of mass.
        """
        return 1 / self.total_mass * np.sum(coords3d * self.masses[:, None], axis=0)

    @property
    def center_of_mass(self):
        """Returns the center of mass.

        Returns
        -------
        R : np.array, shape(3, )
            Center of mass.
        """
        return self.center_of_mass_at(self.coords3d)

    @property
    def centroid(self):
        """Geometric center of the Geometry.

        Returns
        -------
        R : np.array, shape(3, )
            Geometric center of the Geometry.
        """
        return self.coords3d.mean(axis=0)

    def center(self):
        self.coords3d -= self.centroid[None, :]

    @property
    def mw_coords(self):
        """Mass-weighted coordinates.

        Returns
        -------
        mw_coords : np.array
            1d array containing the mass-weighted cartesian coordiantes.
        """
        return np.sqrt(self.masses_rep) * self._coords

    @mw_coords.setter
    def mw_coords(self, mw_coords):
        """Set mass-weighted coordinates."""
        self.coords = mw_coords / np.sqrt(self.masses_rep)

    def fd_coords3d_gen(self, step_size=1e-3):
        """Iterator returning 3d Cartesians for finite-differences."""
        coords3d = self.coords3d.copy()
        zeros = np.zeros_like(coords3d)
        for i, _ in enumerate(self.coords3d):
            for j in (0, 1, 2):
                step = zeros.copy()
                step[i, j] = step_size
                yield i, j, coords3d + step, coords3d - step

    @property
    def covalent_radii(self):
        return np.array([CR[a.lower()] for a in self.atoms])

    @property
    def vdw_radii(self):
        return np.array([VDWR[a.lower()] for a in self.atoms])

    def vdw_volume(self, **kwargs):
        V_au, *_ = molecular_volume(self.coords3d, self.vdw_radii, **kwargs)
        return V_au

    @property
    def inertia_tensor(self):
        return inertia_tensor(self.coords3d, self.masses)

    def principal_axes_are_aligned(self):
        """Check if the principal axes are aligned with the cartesian axes.

        Returns
        -------
        aligned : bool
            Wether the principal axes are aligned or not.
        """
        w, v = np.linalg.eigh(self.inertia_tensor)
        return np.allclose(v, np.eye(3)), v

    def align_principal_axes(self):
        """Align the principal axes to the cartesian axes.

        https://math.stackexchange.com/questions/145023
        """
        I = self.inertia_tensor
        w, v = np.linalg.eigh(I)
        # rot = np.linalg.solve(v, np.eye(3))
        # self.coords3d = rot.dot(self.coords3d.T).T
        self.coords3d = v.T.dot(self.coords3d.T).T

    def standard_orientation(self):
        # Translate center of mass to cartesian origin
        self.coords3d -= self.center_of_mass
        # Try to rotate the principal axes onto the cartesian axes
        for i in range(5):
            self.align_principal_axes()
            aligned, vecs = self.principal_axes_are_aligned()
            if aligned:
                break

    def reparametrize(self):
        if not hasattr(self.calculator, 'get_coords'):
            return False
        try:
            results = self.calculator.get_coords(self.atoms, self.cart_coords)
            self.set_coords(results["coords"], cartesian=True)
            return True
        except Exception:
            return False

    @property
    def energy(self):
        """Energy of the current atomic configuration.

        Returns
        -------
        energy : float
            Energy of the current atomic configuration.
        """
        if self._energy is None:
            results = self.calculator.get_energy(self.atoms, self._coords)
            self.set_results(results)
        return self._energy

    @energy.setter
    def energy(self, energy):
        """Internal wrapper for setting the energy.

        Parameters
        ----------
        energy : float
        """
        self._energy = energy

    @property
    def all_energies(self):
        """Return energies of all states that were calculated.

        This will also set self.energy, which may NOT be the ground state,
        but the state correspondig to the 'root' attribute of the calculator."""
        if self._all_energies is None:
            results = self.calculator.get_energy(self.atoms, self._coords)
            self.set_results(results)
        return self._all_energies

    @all_energies.setter
    def all_energies(self, all_energies):
        """Internal wrapper for setting all energies.

        Parameters
        ----------
        all_energies : np.array
        """
        self._all_energies = all_energies

    @property
    def cart_forces(self):
        if self._forces is None:
            results = self.calculator.get_forces(self.atoms, self._coords)
            self.set_results(results)
        return self._forces

    @cart_forces.setter
    def cart_forces(self, cart_forces):
        cart_forces = np.array(cart_forces)
        assert cart_forces.shape == self.cart_coords.shape
        self._forces = cart_forces

    @property
    def forces(self):
        """Energy of the current atomic configuration.

        Returns
        -------
        force : np.array
            1d array containing the forces acting on the atoms. Negative
            of the gradient.
        """
        forces = self.cart_forces
        if self.internal:
            forces = self.internal.transform_forces(forces)
        return forces

    @forces.setter
    def forces(self, forces):
        """Internal wrapper for setting the forces.

        Parameters
        ----------
        forces : np.array
        """
        forces = np.array(forces)
        assert forces.shape == self.cart_coords.shape
        self._forces = forces

    @property
    def cart_gradient(self):
        return -self.cart_forces

    @cart_gradient.setter
    def cart_gradient(self, cart_gradient):
        self.cart_forces = -cart_gradient

    @property
    def gradient(self):
        """Negative of the force.

        Returns
        -------
        gradient : np.array
            1d array containing the negative of the current forces.
        """
        return -self.forces

    # @gradient.setter
    # def gradient(self, gradient):
    # """Internal wrapper for setting the gradient."""
    # # No check here as this is handled by in the forces.setter.
    # self.forces = -gradient

    @property
    def mw_gradient(self):
        """Mass-weighted gradient.

        Returns
        -------
        mw_gradient : np.array
            Returns the mass-weighted gradient.
        """
        return -self.forces / np.sqrt(self.masses_rep)

    @property
    def cart_hessian(self):
        if self._hessian is None:
            results = self.calculator.get_hessian(self.atoms, self._coords)
            self.set_results(results)
        return self._hessian

    @cart_hessian.setter
    def cart_hessian(self, cart_hessian):
        if cart_hessian is not None:
            # cart_hessian = np.array(cart_hessian)
            if self.within_partial_hessian is not None:
                active_n_dof = int(self.within_partial_hessian.get("active_n_dof", 0))
                full_n_dof = int(
                    self.within_partial_hessian.get("full_n_dof", self.cart_coords.size)
                )
                if cart_hessian.shape != (full_n_dof, full_n_dof):
                    assert cart_hessian.shape == (active_n_dof, active_n_dof)
            else:
                assert cart_hessian.shape == (self.cart_coords.size, self.cart_coords.size)
        self._hessian = cart_hessian

    @property
    def hessian(self):
        """Matrix of second derivatives of the energy in respect to atomic
        displacements.

        Returns
        -------
        hessian : np.array
            2d array containing the second derivatives of the energy with respect
            to atomic/coordinate displacements depending on the type of
            coordiante system.
        """
        hessian = self.cart_hessian
        if self.internal:
            int_gradient = self.gradient
            return self.internal.transform_hessian(hessian, int_gradient)
        return hessian

    # @hessian.setter
    # def hessian(self, hessian):
    # """Internal wrapper for setting the hessian."""
    # assert hessian.shape == (self.coords.size, self.coords.size)
    # self._hessian = hessian

    def mass_weigh_hessian(self, hessian):
        if (
            self.within_partial_hessian is not None
            and hessian is not None
            and hessian.shape == (int(self.within_partial_hessian.get("active_n_dof", 0)),
                                  int(self.within_partial_hessian.get("active_n_dof", 0)))
        ):
            act_atoms = self.hess_active_atom_indices
            masses_act = self.masses[act_atoms]
            m3 = np.repeat(masses_act, 3)
            inv_sqrt_m = 1.0 / np.sqrt(m3)
            if isinstance(hessian, torch.Tensor):
                inv = torch.as_tensor(inv_sqrt_m, dtype=hessian.dtype, device=hessian.device)
                hessian.mul_(inv.view(-1, 1))
                hessian.mul_(inv.view(1, -1))
                return hessian
            hessian *= inv_sqrt_m[:, None]
            hessian *= inv_sqrt_m[None, :]
            return hessian

        inv_sqrt_m = 1.0 / (self.masses_rep ** 0.5)
        if isinstance(hessian, torch.Tensor):
            s = torch.tensor(inv_sqrt_m, dtype=hessian.dtype, device=hessian.device)
            return hessian * s[:, None] * s[None, :]
        else:
            return hessian * inv_sqrt_m[:, None] * inv_sqrt_m[None, :]

    @property
    def mw_hessian(self):
        """Mass-weighted hessian.

        Returns
        -------
        mw_hessian : np.array
            2d array containing the mass-weighted hessian M^(-1/2) H M^(-1/2).
        """
        # M^(-1/2) H M^(-1/2)
        # TODO: Do the right thing here when the hessian is not yet calculated.
        #       this would probably involve figuring out how to mass-weigh and
        #       internal coordinat hessian... I think this is described in one
        #       of the Gonzalez-Schlegel-papers about the GS2 algorithm.
        return self.mass_weigh_hessian(self.cart_hessian)

    def unweight_mw_hessian(self, mw_hessian):
        """Unweight a mass-weighted hessian.

        Parameters
        ----------
        mw_hessian : np.array
            Mass-weighted hessian to be unweighted.

        Returns
        -------
        hessian : np.array
            2d array containing the hessian.
        """
        sqrt_m = self.masses_rep ** 0.5
        if isinstance(mw_hessian, torch.Tensor):
            s = torch.tensor(sqrt_m, dtype=mw_hessian.dtype, device=mw_hessian.device)
            return mw_hessian * s[:, None] * s[None, :]
        else:
            return mw_hessian * sqrt_m[:, None] * sqrt_m[None, :]

    # indices (0 … N‑1) of atoms that are *not* frozen
    @property
    def active_atom_indices(self):
        if not hasattr(self, "_active_atom_indices"):
            self._active_atom_indices = np.array(
                [i for i in range(len(self.atoms)) if i not in self.freeze_atoms],
                dtype=int,
            )
        return self._active_atom_indices

    # 3N‑dimensional indices of DOFs that are not frozen (x,y,z per atom)
    @property
    def active_dof_indices(self):
        if not hasattr(self, "_active_dof_indices"):
            act = []
            for a in self.active_atom_indices:
                act.extend([3 * a, 3 * a + 1, 3 * a + 2])
            self._active_dof_indices = np.asarray(act, dtype=int)
        return self._active_dof_indices

    @property
    def hess_active_atom_indices(self):
        if self.within_partial_hessian is None:
            return self.active_atom_indices
        return self.within_partial_hessian["active_atoms"]

    @property
    def hess_active_dof_indices(self):
        if self.within_partial_hessian is None:
            return self.active_dof_indices
        return self.within_partial_hessian["active_dofs"]

    # Convenience: extract / insert an active slice
    def full_from_active(self, active_vec):
        """Expand a vector defined on active DOFs to 3N, keeping frozen data."""
        if isinstance(active_vec, torch.Tensor):
            full = torch.zeros_like(self.cart_coords)
            full[self.active_dof_indices] = active_vec
            return full
        full = np.zeros_like(self.cart_coords)
        full[self.active_dof_indices] = active_vec
        return full

    def active_from_full(self, full_vec):
        """Return the part of a 3N vector that belongs to active DOFs."""
        return full_vec[self.active_dof_indices]

    def full_from_hess_active(self, active_vec):
        """Expand a vector defined on Hessian-active DOFs to full 3N."""
        inds = self.hess_active_dof_indices
        if isinstance(active_vec, torch.Tensor):
            idx = torch.as_tensor(inds, dtype=torch.long, device=active_vec.device)
            full = torch.zeros(self.cart_coords.size, dtype=active_vec.dtype, device=active_vec.device)
            full.index_copy_(0, idx, active_vec)
            return full
        full = np.zeros_like(self.cart_coords)
        full[inds] = active_vec
        return full

    def hess_active_from_full(self, full_vec):
        """Return the part of a 3N vector that belongs to Hessian-active DOFs."""
        inds = self.hess_active_dof_indices
        if isinstance(full_vec, torch.Tensor):
            idx = torch.as_tensor(inds, dtype=torch.long, device=full_vec.device)
            return full_vec.index_select(0, idx)
        return full_vec[inds]

    def set_h5_hessian(self, fn):
        with h5py.File(fn, "r") as handle:
            atoms = handle.attrs["atoms"]
            hessian = handle["hessian"][:]

        # Also check lengths, as zip would lead to trunction for
        # different lenghts of self.atoms and atoms.
        valid = (len(atoms) == len(self.atoms)) and all(
            [ga.lower() == a.lower() for ga, a in zip(self.atoms, atoms)]
        )
        if valid:
            self.cart_hessian = hessian

    def get_normal_modes(self, cart_hessian=None, full=False):
        """Normal mode wavenumbers, eigenvalues and Cartesian displacements Hessian."""
        if cart_hessian is None:
            cart_hessian = self.cart_hessian

        mw_hessian = self.mass_weigh_hessian(cart_hessian)
        proj_hessian, P = self.eckart_projection(mw_hessian, return_P=True, full=full)

        is_partial = (
            self.within_partial_hessian is not None
            and proj_hessian is not None
            and proj_hessian.shape == (int(self.within_partial_hessian.get("active_n_dof", 0)),
                                       int(self.within_partial_hessian.get("active_n_dof", 0)))
        )

        if isinstance(proj_hessian, torch.Tensor):
            eigvals, eigvecs = torch.linalg.eigh(proj_hessian)
            mw_cart_displs = P.T @ eigvecs
            if is_partial:
                masses_act = self.masses[self.hess_active_atom_indices]
                m3 = torch.repeat_interleave(
                    torch.as_tensor(masses_act, dtype=proj_hessian.dtype, device=proj_hessian.device), 3
                )
                cart_displs_act = mw_cart_displs / torch.sqrt(m3.view(-1, 1))
                cart_displs_act /= torch.linalg.norm(cart_displs_act, dim=0)
                cart_displs = torch.zeros(
                    (self.cart_coords.size, cart_displs_act.shape[1]),
                    dtype=cart_displs_act.dtype,
                    device=cart_displs_act.device,
                )
                idx = torch.as_tensor(self.hess_active_dof_indices, dtype=torch.long, device=cart_displs.device)
                cart_displs.index_copy_(0, idx, cart_displs_act)
            else:
                inv_sqrt_m = torch.tensor(1.0 / (self.masses_rep ** 0.5), dtype=proj_hessian.dtype, device=proj_hessian.device)
                cart_displs = mw_cart_displs * inv_sqrt_m[:, None]
                cart_displs /= torch.linalg.norm(cart_displs, dim=0)
            eigvals = eigvals.cpu().numpy()
        else:
            eigvals, eigvecs = np.linalg.eigh(proj_hessian)
            mw_cart_displs = P.T.dot(eigvecs)
            if is_partial:
                masses_act = self.masses[self.hess_active_atom_indices]
                m3 = np.repeat(masses_act, 3)
                cart_displs_act = mw_cart_displs / np.sqrt(m3)[:, None]
                cart_displs_act /= np.linalg.norm(cart_displs_act, axis=0)
                cart_displs = np.zeros((self.cart_coords.size, cart_displs_act.shape[1]))
                cart_displs[self.hess_active_dof_indices, :] = cart_displs_act
            else:
                inv_sqrt_m = 1.0 / (self.masses_rep ** 0.5)
                cart_displs = mw_cart_displs * inv_sqrt_m[:, None]
                cart_displs /= np.linalg.norm(cart_displs, axis=0)

        nus = eigval_to_wavenumber(eigvals)
        return nus, eigvals, mw_cart_displs, cart_displs

    def get_imag_frequencies(self, hessian=None, thresh=1e-6):
        vibfreqs, eigvals, *_ = self.get_normal_modes(hessian)
        return vibfreqs[eigvals < thresh]

    def get_thermoanalysis(
        self, energy=None, cart_hessian=None, T=T_DEFAULT, p=p_DEFAULT, point_group="c1"
    ):
        if cart_hessian is None:
            cart_hessian = self.cart_hessian
            # Delte any supplied energy value when a Hessian calculation is carried out
            energy = None

        if energy is None:
            energy = self.energy

        vibfreqs, *_ = self.get_normal_modes(cart_hessian)
        try:
            mult = self.calculator.mult
        except AttributeError:
            mult = 1
            logger.debug(
                "Multiplicity for electronic entropy could not be determined! "
                f"Using 2S+1 = {mult}."
            )

        thermo_dict = {
            "masses": self.masses,
            "wavenumbers": vibfreqs,
            "coords3d": self.coords3d,
            "scf_energy": energy,
            "mult": mult,
        }

        qcd = QCData(thermo_dict, point_group=point_group)
        thermo = thermochemistry(
            qcd, temperature=T, pressure=p, invert_imags=-15.0, cutoff=25.0
        )

        return thermo

    def get_trans_rot_projector(self, full=False):
        return get_trans_rot_projector(self.cart_coords, masses=self.masses, full=full)

    def eckart_projection(self, mw_hessian, return_P=False, full=False):
        # Must not project analytical 2d potentials.
        if self.is_analytical_2d:
            return mw_hessian

        if (
            self.within_partial_hessian is not None
            and mw_hessian is not None
            and mw_hessian.shape == (int(self.within_partial_hessian.get("active_n_dof", 0)),
                                     int(self.within_partial_hessian.get("active_n_dof", 0)))
        ):
            coords_act = self.coords3d[self.hess_active_atom_indices].flatten()
            masses_act = self.masses[self.hess_active_atom_indices]
            P = get_trans_rot_projector(coords_act, masses=masses_act, full=full)
        else:
            P = self.get_trans_rot_projector(full=full)
        if isinstance(mw_hessian, torch.Tensor):
            P = torch.tensor(P, device=mw_hessian.device, dtype=mw_hessian.dtype)
            proj_hessian = P @ mw_hessian @ P.T
            # Projection seems to slightly break symmetry (sometimes?). Resymmetrize.
            proj_hessian = (proj_hessian + proj_hessian.T) / 2
        else:
            proj_hessian = P.dot(mw_hessian).dot(P.T)
            # Projection seems to slightly break symmetry (sometimes?). Resymmetrize.
            proj_hessian = (proj_hessian + proj_hessian.T) / 2
        if return_P:
            return proj_hessian, P
        else:
            return proj_hessian

    def calc_energy_and_forces(self):
        """Force a calculation of the current energy and forces."""
        results = self.calculator.get_forces(self.atoms, self.cart_coords)
        self.set_results(results)

    def assert_cart_coords(self, coords):
        assert coords.size == self.cart_coords.size, (
            "This method only works with cartesian coordinate input. "
            "Did you accidentally provide internal coordinates?"
        )

    def get_temporary_coords(self, coords):
        if self.coord_type != "cart":
            int_step = coords - self.internal.coords
            cart_step = self.internal.transform_int_step(int_step, pure=True)
            coords = self.cart_coords + cart_step
        self.assert_cart_coords(coords)
        return coords

    def get_energy_at(self, coords):
        coords = self.get_temporary_coords(coords)
        return self.calculator.get_energy(self.atoms, coords)["energy"]

    def get_energy_at_cart_coords(self, cart_coords):
        self.assert_cart_coords(cart_coords)
        return self.calculator.get_energy(self.atoms, cart_coords)["energy"]

    def get_energy_and_forces_at(self, coords):
        """Calculate forces and energies at the given coordinates.

        The results are not saved in the Geometry object."""
        coords = self.get_temporary_coords(coords)
        results = self.calculator.get_forces(self.atoms, coords)
        self.zero_frozen_forces(results["forces"])

        if self.coord_type != "cart":
            results["forces"] = self.internal.transform_forces(results["forces"])

        return results

    def get_energy_and_cart_forces_at(self, cart_coords):
        self.assert_cart_coords(cart_coords)
        results = self.calculator.get_forces(self.atoms, cart_coords)
        self.zero_frozen_forces(results["forces"])
        return results

    def get_energy_and_cart_hessian_at(self, cart_coords):
        self.assert_cart_coords(cart_coords)
        results = self.calculator.get_hessian(self.atoms, cart_coords)
        return results

    def calc_double_ao_overlap(self, geom2):
        return self.calculator.run_double_mol_calculation(
            self.atoms, self.coords, geom2.coords
        )

    def zero_frozen_forces(self, cart_forces):
        cart_forces.reshape(-1, 3)[self.freeze_atoms] = 0.0

    def clear(self):
        """Reset the object state."""

        self._energy = None
        self._forces = None
        self._hessian = None
        self.within_partial_hessian = None
        self.true_energy = None
        self.true_forces = None
        self.true_hessian = None
        self._all_energies = None
        self.results = {}

    def set_results(self, results):
        """Save the results from a dictionary.

        Parameters
        ----------
        results : dict
            The keys in this dict will be set as attributes in the current
            object, with the corresponding item as value.
        """

        if "within_partial_hessian" in results:
            self.within_partial_hessian = results["within_partial_hessian"]
        elif "hessian" in results:
            self.within_partial_hessian = None

        trans = {
            "energy": "energy",
            "forces": "cart_forces",
            "hessian": "cart_hessian",
            "within_partial_hessian": "within_partial_hessian",
            # True properties in AFIR calculations
            "true_energy": "true_energy",
            "true_forces": "true_forces",
            "true_hessian": "true_hessian",
            # Overlap calculator; includes excited states
            "all_energies": "all_energies",
        }

        for key in results:
            if key == "within_partial_hessian":
                continue
            # Zero forces of frozen atoms
            if key == "forces":
                self.zero_frozen_forces(results[key])

            setattr(self, trans[key], results[key])
        self.results = results

    def as_xyz(self, comment="", atoms=None, cart_coords=None):
        """Current geometry as a string in XYZ-format.

        Parameters
        ----------
        comment : str, optional
            Will be written in the second line (comment line) of the
            XYZ-string.
        cart_coords : np.array, 1d, shape (3 * atoms.size, )
            Cartesians for dumping instead of self._coords.

        Returns
        -------
        xyz_str : str
            Current geometry as string in XYZ-format.
        """
        if atoms is None:
            atoms = self.atoms
        if cart_coords is None:
            cart_coords = self._coords
        cart_coords = cart_coords.copy()
        cart_coords *= BOHR2ANG
        if comment == "":
            comment = self.comment
        return make_xyz_str(atoms, cart_coords.reshape((-1, 3)), comment)

    def dump_xyz(self, fn, cart_coords=None, **kwargs):
        fn = str(fn)
        if not fn.lower().endswith(".xyz"):
            fn = fn + ".xyz"
        with open(fn, "w") as handle:
            handle.write(self.as_xyz(cart_coords=cart_coords, **kwargs))

    def get_subgeom(self, indices, coord_type="cart", sort=False):
        """Return a Geometry containing a subset of the current Geometry.

        Parameters
        ----------
        indices : iterable of ints
            Atomic indices that the define the subset of the current Geometry.
        coord_type : str, ("cart", "redund"), optional
            Coordinate system of the new Geometry.

        Returns
        -------
        sub_geom : Geometry
            Subset of the current Geometry.
        """
        if sort:
            indices = sorted(indices)
        ind_list = list(indices)
        sub_atoms = [self.atoms[i] for i in ind_list]
        sub_coords = self.coords3d[ind_list]
        sub_geom = Geometry(sub_atoms, sub_coords.flatten(), coord_type=coord_type)
        return sub_geom

    def get_subgeom_without(self, indices, **kwargs):
        with_indices = [ind for ind, _ in enumerate(self.atoms) if ind not in indices]
        return self.get_subgeom(with_indices, **kwargs)

    def rmsd(self, geom):
        return rmsd.kabsch_rmsd(
            self.coords3d - self.centroid, geom.coords3d - geom.centroid
        )

    def as_g98_list(self):
        """Returns data for fake Gaussian98 standard orientation output.

        Returns
        -------
        g98_list : list
            List with one row per atom. Every row contains [center number,
            atomic number, atomic type (always 0 for now), X Y Z coordinates
            in Angstrom.
        """
        Atom = namedtuple("Atom", "center_num atom_num atom_type x y z")
        atoms = list()
        for i, (a, c) in enumerate(zip(self.atoms, self.coords3d), 1):
            x, y, z = c * BOHR2ANG
            atom = Atom(i, ATOMIC_NUMBERS[a.lower()], 0, x, y, z)
            atoms.append(atom)
        return atoms

    def tmp_xyz_handle(self, atoms=None, cart_coords=None):
        tmp_xyz = tempfile.NamedTemporaryFile(suffix=".xyz")
        tmp_xyz.write(self.as_xyz(atoms=atoms, cart_coords=cart_coords).encode("utf-8"))
        tmp_xyz.flush()
        return tmp_xyz

    def jmol(self, atoms=None, cart_coords=None):
        """Show geometry in jmol."""
        tmp_xyz = self.tmp_xyz_handle(atoms, cart_coords)
        jmol_cmd = "jmol"
        try:
            subprocess.run([jmol_cmd, tmp_xyz.name])
        except FileNotFoundError:
            print(f"'{jmol_cmd}' seems not to be on your path!")
        tmp_xyz.close()

    def modes3d(self):
        try:
            bonds = self.internal.bond_atom_indices
            bonds_str = " --bonds " + " ".join(map(str, it.chain(*bonds)))
        except AttributeError:
            bonds_str = ""

        tmp_xyz = self.tmp_xyz_handle()
        cmd = ["modes3d.py", tmp_xyz.name]
        if bonds_str:
            cmd.extend(shlex.split(bonds_str))
        subprocess.run(cmd)
        tmp_xyz.close()

    def as_ase_atoms(self):
        try:
            import ase
        except ImportError:
            print("Please install the 'ase' package!")
            return None

        # ASE coordinates are in Angstrom
        atoms = ase.Atoms(symbols=self.atoms, positions=self.coords3d * BOHR2ANG)

        if self.calculator is not None:
            from pysisyphus.calculators import FakeASE

            ase_calc = FakeASE(self.calculator)
            atoms.set_calculator(ase_calc)
        return atoms

    def get_restart_info(self):
        # Geometry restart information
        restart_info = {
            "atoms": self.atoms,
            "cart_coords": self.cart_coords.tolist(),
            "coord_type": self.coord_type,
            "comment": self.comment,
        }
        try:
            typed_prims = self.internal.typed_prims
        except AttributeError:
            typed_prims = None
        restart_info["typed_prims"] = typed_prims

        # Calculator restart information
        try:
            calc_restart_info = self.calculator.get_restart_info()
        except AttributeError:
            calc_restart_info = dict()
        restart_info["calc_info"] = calc_restart_info

        return restart_info

    def set_restart_info(self, restart_info):
        assert self.atoms == restart_info["atoms"]
        self.cart_coords = np.array(restart_info["cart_coords"], dtype=float)

        try:
            self.calculator.set_restart_info(restart_info["calc_info"])
        except KeyError:
            print("No calculator restart information found!")
        except AttributeError:
            print("Could not restart calculator, as no calculator is set!")

    def get_sphere_radius(self, offset=4):
        distances = pdist(self.coords3d)

        radius = (distances.max() / 2) + offset
        return radius

    def without_hydrogens(self):
        atoms_no_h, coords3d_no_h = zip(
            *[
                (atom, coords)
                for atom, coords in zip(self.atoms, self.coords3d)
                if atom.lower() != "h"
            ]
        )
        return Geometry(atoms_no_h, np.array(coords3d_no_h).flatten())

    def describe(self):
        return f"Geometry({self.sum_formula}, {len(self.atoms)} atoms)"

    def approximate_radius(self):
        """Approximate molecule radius from the biggest atom distance along an axis."""
        coords3d = self.coords3d - self.centroid[None, :]
        mins = coords3d.min(axis=0)
        maxs = coords3d.max(axis=0)
        dists = maxs - mins
        max_dist = dists.max()
        return max_dist

    def rotate(self, copy=False, rng=None):
        if copy:
            geom = self.copy()
        else:
            geom = self

        rot = Rotation.random(random_state=rng)
        geom.coords3d = rot.apply(geom.coords3d)
        return geom

    @property
    def bond_sets(self, bond_factor=BOND_FACTOR):
        bonds = find_bonds(
            self.atoms, self.coords3d, self.covalent_radii, bond_factor=bond_factor
        )
        bond_sets = set([frozenset(b) for b in bonds])
        return bond_sets

    def __str__(self):
        name = ""
        if self.name:
            name = f"{self.name}, "
        return f"Geometry({name}{self.sum_formula})"

    def __repr__(self):
        return self.__str__()



from math import ceil
from pathlib import Path
import re
import textwrap

import jinja2
import numpy as np

from pysisyphus.calculators.OverlapCalculator import OverlapCalculator
from pysisyphus.constants import BOHR2ANG, AU2EV


def parse_mo(eigvec):
    coeffs = list()
    # Skip first line
    for line in eigvec.split("\n")[1:]:
        line = line.strip()
        if line == "":
            continue
        line = line.split()
        if len(line) == 5:
            line = line[-3:]
        coeffs.append(line[1])
    return coeffs


def parse_xplusy(text):
    lines = text.split("\n")
    size, states = [int(i) for i in lines.pop(0).split()]
    # Add one for header line
    block_size = ceil(size / 6) + 1
    assert len(lines) == (states * block_size)

    xpys = list()
    for i in range(states):
        block = lines[i * block_size : (i + 1) * block_size]
        _, *rest = block
        xpy = np.array([line.split() for line in rest], dtype=float)
        xpys.append(xpy)
    return size, states, np.array(xpys)


class DFTBp(OverlapCalculator):

    conf_key = "dftbp"
    _set_plans = (
        "out",
        "exc_dat",
        "eigenvec",
        "xplusy_dat",
    )

    max_ang_moms = {
        "mio-ext": {
            "H": "s",
            "C": "p",
            "N": "p",
            "O": "p",
        },
        "3ob": {
            "Br": "d",
            "C": "p",
            "Ca": "p",
            "Cl": "d",
            "F": "p",
            "H": "s",
            "I": "d",
            "K": "p",
            "Mg": "p",
            "N": "p",
            "Na": "p",
            "O": "p",
            "P": "d",
            "S": "d",
            "Zn": "d",
        },
    }
    hubbard_derivs = {
        "3ob": {
            "Br": -0.0573,
            "Mg": -0.0200,
            "C": -0.1492,
            "N": -0.1535,
            "Ca": -0.0340,
            "Na": -0.0454,
            "Cl": -0.0697,
            "O": -0.1575,
            "F": -0.1623,
            "P": -0.1400,
            "H": -0.1857,
            "S": -0.1100,
            "I": -0.0433,
            "Zn": -0.0300,
            "K": -0.0339,
        },
    }

    def __init__(self, parameter, *args, slakos=None, root=None, **kwargs):
        super().__init__(*args, **kwargs)

        assert self.mult == 1, "Open-shell not yet supported!"
        self.parameter = parameter
        if slakos is None:
            slakos = self.get_cmd("slakos")
        self.slakos_prefix = str(slakos)
        assert (Path(self.slakos_prefix) / self.parameter).exists(), (
            f"Expected '{self.parameter}' sub-directory in '{self.slakos_prefix}' "
            "but could not find it!"
        )
        self.root = root

        self.base_cmd = self.get_cmd()
        self.gen_geom_fn = "geometry.gen"
        self.inp_fn = "dftb_in.hsd"
        self.out_fn = "dftb.out"
        self.to_keep = (
            "dftb_in.hsd",
            "out:detailed.out",
            self.gen_geom_fn,
            self.out_fn,
            "xplusy_dat:XplusY.DAT",
            "exc_dat:EXC.DAT",
            "eigenvec:eigenvec.out",
        )
        self.parser_funcs = {
            "energy": self.parse_energy,
            "forces": self.parse_forces,
        }

        self.dftb_tpl = jinja2.Template(
            textwrap.dedent(
                """
            Geometry = GenFormat {
              <<< "{{ gen_geom_fn }}"
            }
            Hamiltonian = DFTB {
              Scc = Yes
              Charge = {{ charge }}
              SpinPolarisation = {{ spinpol }}
              SlaterKosterFiles = Type2FileNames {
                  Prefix = "{{ slakos_prefix }}/{{ parameter }}/"
                  Separator = "-"
                  Suffix = ".skf"
                }
              MaxAngularMomentum {
               {%- for atom, ang_mom in max_ang_moms %}
                {{ atom }} = "{{ ang_mom }}"
               {%- endfor %}
              }
              {% if parameter == "3ob" %}
                ThirdOrderFull = Yes
                HubbardDerivs {
                {% for atom, hd in hubbard_derivs -%}
                 {{ atom }} = {{ hd }}
                {% endfor %}
                }
                HCorrection = Damping {
                 Exponent = 4.00
                }
              {% endif %}
            }

            ExcitedState {
             {{ excited_state_str }}
            }

            Analysis {
             {% for anal in analysis %}
              {{ anal }}
             {%- endfor %}
            }

            ParserOptions {
              ParserVersion = 8
            }
        """
            ).strip()
        )

    @staticmethod
    def get_gen_str(atoms, coords):
        gen_fmt_tpl = jinja2.Template(
            textwrap.dedent(
                """
            {{ atom_num }} C
            {% for atom in atom_types %}{{ atom }} {% endfor %}
            {% for type_, x, y, z in types_xyz %}
             {{ loop.index }} {{type_}}  {{x}} {{y}} {{z}}
            {%- endfor %}
        """
            ).strip()
        )
        atom_num = len(atoms)
        unique_atoms = tuple(set(atoms))
        atom_types = {atom: i for i, atom in enumerate(unique_atoms, 1)}
        c3d = coords.reshape(-1, 3) * BOHR2ANG
        types_xyz = [(atom_types[atom], x, y, z) for atom, (x, y, z) in zip(atoms, c3d)]
        gen_str = gen_fmt_tpl.render(
            atom_num=atom_num,
            atom_types=unique_atoms,
            types_xyz=types_xyz,
        )
        return gen_str

    @staticmethod
    def get_excited_state_str(root, forces=False):
        if root is None:
            return ""

        casida_tpl = jinja2.Template(
            textwrap.dedent(
                """
            Casida {
                NrOfExcitations = {{ nstates }}
                Symmetry = Singlet
                StateOfInterest = {{ root }}
                WriteXplusY = Yes
                {{ es_forces }}
            }
            """
            )
        )
        es_forces = "ExcitedStateForces = Yes" if forces else ""
        es_str = casida_tpl.render(
            nstates=root + 5,
            root=root,
            es_forces=es_forces,
        )
        return es_str

    def prepare_input(self, atoms, coords, calc_type):
        path = self.prepare_path(use_in_run=True)
        gen_str = self.get_gen_str(atoms, coords)
        with open(path / self.gen_geom_fn, "w") as handle:
            handle.write(gen_str)
        analysis = list()
        if calc_type == "forces":
            analysis.append("CalculateForces = Yes")
        if self.root:
            analysis.extend(("WriteEigenvectors = Yes", "EigenvectorsAsText = Yes"))
        ang_moms = self.max_ang_moms[self.parameter]
        unique_atoms = set(atoms)
        max_ang_moms = [(atom, ang_moms[atom]) for atom in unique_atoms]

        # spinpol = (
        # "{}"
        # if (self.mult == 1)
        # else f"Colinear {{ UnpairedElectrons = {self.mult-1} }}"
        # )
        spinpol = "{}"
        es_forces = calc_type == "forces"
        try:
            param_hubbard_derivs = self.hubbard_derivs[self.parameter]
            hubbard_derivs = [
                (atom, param_hubbard_derivs[atom]) for atom in unique_atoms
            ]
        except KeyError:
            hubbard_derivs = list()

        inp = self.dftb_tpl.render(
            gen_geom_fn=self.gen_geom_fn,
            charge=self.charge,
            spinpol=spinpol,
            slakos_prefix=self.slakos_prefix,
            parameter=self.parameter,
            max_ang_moms=max_ang_moms,
            hubbard_derivs=hubbard_derivs,
            excited_state_str=self.get_excited_state_str(self.root, es_forces),
            analysis=analysis,
        )
        return inp, path

    def store_and_track(self, results, func, atoms, coords, **prepare_kwargs):
        if self.track:
            self.store_overlap_data(atoms, coords)
            if self.track_root():
                # Redo the calculation with the updated root
                results = func(atoms, coords, **prepare_kwargs)
        results["all_energies"] = self.parse_all_energies()
        return results

    def get_energy(self, atoms, coords, **prepare_kwargs):
        inp, path = self.prepare_input(atoms, coords, "energy")
        results = self.run(inp, "energy")
        results = self.store_and_track(
            results, self.get_energy, atoms, coords, **prepare_kwargs
        )
        return results

    def get_forces(self, atoms, coords, **prepare_kwargs):
        inp, path = self.prepare_input(atoms, coords, "forces")
        run_kwargs = {
            "calc": "forces",
            "hold": self.track,
        }
        results = self.run(inp, **run_kwargs)
        results = self.store_and_track(
            results, self.get_forces, atoms, coords, **prepare_kwargs
        )
        # if self.track:
        # self.calc_counter += 1
        # self.store_overlap_data(atoms, coords, path)
        # if self.track_root():
        # # Redo the calculation with the updated root
        # results = self.get_forces(atoms, coords)
        # try:
        # shutil.rmtree(path)
        # except FileNotFoundError:
        # self.log(f"'{path}' has already been deleted!")
        return results

    def run_calculation(self, atoms, coords, **prepare_kwargs):
        inp, path = self.prepare_input(atoms, coords, "energy")
        run_kwargs = {
            "calc": "energy",
            "hold": self.track,
        }
        results = self.run(inp, **run_kwargs)
        if self.track:
            self.calc_counter += 1
            self.store_overlap_data(atoms, coords, path)
        return results

    def parse_total_energy(self, text):
        energy_re = re.compile(r"Total energy:\s*([-\d\.]+)\s*H")
        exc_energy_re = re.compile(r"Excitation Energy:\s*([\-\.\d]+)\s*H")
        energy = float(energy_re.search(text)[1])
        exc_mobj = exc_energy_re.search(text)
        if exc_mobj:
            energy += float(exc_mobj[1])
        return energy

    def parse_energy(self, path):
        detailed = path / "detailed.out"
        with open(detailed) as handle:
            text = handle.read()
        results = {
            "energy": self.parse_total_energy(text),
        }
        return results

    @staticmethod
    def parse_exc_dat(text):
        exc_re = re.compile("=+(.+)", re.DOTALL)
        mobj = exc_re.search(text)
        exc_lines = mobj[1].strip().split("\n")
        exc_ens = (
            np.array([line.strip().split()[0] for line in exc_lines], dtype=float)
            / AU2EV
        )
        return exc_ens

    def parse_forces(self, path):
        forces_re = re.compile("Total Forces(.+)Maximal derivative", re.DOTALL)
        detailed = path / "detailed.out"
        with open(detailed) as handle:
            text = handle.read()
        mobj = forces_re.search(text)
        forces = np.array(mobj[1].strip().split(), dtype=float).reshape(-1, 4)[:, 1:]
        results = {
            "energy": self.parse_total_energy(text),
            "forces": forces.flatten(),
        }
        return results

    def parse_all_energies(self, out_fn=None, exc_dat=None):
        if out_fn is None:
            out_fn = self.out
        if exc_dat is None:
            try:
                exc_dat = self.exc_dat
            except AttributeError:
                exc_dat = None

        with open(out_fn) as handle:
            detailed = handle.read()

        gs_energy = self.parse_total_energy(detailed)
        if (exc_dat is not None) and exc_dat.exists():
            with open(exc_dat) as handle:
                exc_dat = handle.read()
            exc_ens = self.parse_exc_dat(exc_dat)
            all_energies = np.full(len(exc_ens) + 1, gs_energy)
            all_energies[1:] += exc_ens
        else:
            all_energies = np.array((gs_energy,))
        return all_energies

    def prepare_overlap_data(self, path):
        #
        # Excitation energies
        #
        # import pdb; pdb.set_trace()  # fmt: skip
        # with open(path / "detailed.out") as handle:
        with open(self.out) as handle:
            detailed = handle.read()
        all_energies = self.parse_all_energies(path)

        #
        # MO coefficients
        #
        # with open(path / "eigenvec.out") as handle:
        with open(self.eigenvec) as handle:
            eigenvecs = handle.read().strip()
        eigenvecs = eigenvecs.split("Eigenvector")[1:]

        C = np.array([parse_mo(eigvec) for eigvec in eigenvecs], dtype=float).T
        assert C.shape[0] == C.shape[1]

        #
        # CI coefficients
        #
        electron_re = re.compile(r"Nr. of electrons \(up\):\s*([\d\.]+)")
        mobj = electron_re.search(detailed)
        electrons = int(float(mobj[1]))
        assert electrons % 2 == 0
        occ = electrons // 2
        mo_num = C.shape[0]
        vir = mo_num - occ
        # X+Y
        with open(self.xplusy_dat) as handle:
            xpy_text = handle.read().strip()
        size, states, xpy = parse_xplusy(xpy_text)
        # As of DFTB+ 22.2, we can only get X+Y, but not X-Y, so we can't reconstruct
        # the true X and Y vectors.
        XpY = xpy.reshape(states, occ, vir)
        X = XpY
        Y = np.zeros_like(XpY)
        assert size == occ * vir
        return C, X, Y, all_energies



import glob
import io
from math import sqrt
from pathlib import Path
import os
import re
import struct
import shutil
import warnings

import numpy as np
import pyparsing as pp

from pysisyphus.calculators.OverlapCalculator import OverlapCalculator
from pysisyphus.constants import BOHR2ANG, ANG2BOHR
from pysisyphus.helpers_pure import file_or_str
from pysisyphus.wavefunction import norm_ci_coeffs, Wavefunction


def make_sym_mat(table_block):
    mat_size = int(table_block[1])
    # Orca prints blocks of 5 columns
    arr = np.array(table_block[2:], dtype=float)
    assert arr.size == mat_size**2
    block_size = 5 * mat_size
    cbs = [
        arr[i * block_size : (i + 1) * block_size].reshape(mat_size, -1)
        for i in range(arr.size // block_size + 1)
    ]
    return np.concatenate(cbs, axis=1)


def save_orca_pc_file(point_charges, pc_fn, hardness=None):
    point_charges = point_charges.copy()
    # ORCA excepcts point charge positions in Angstrom
    point_charges[:, :3] *= BOHR2ANG

    # ORCA also expects the ordering <q> <x> <y> <z>, so we have to resort.
    shape = point_charges.shape
    if hardness is not None:
        shape = shape[0], shape[1] + 1
    point_charges_orca = np.zeros_like(point_charges)
    point_charges_orca = np.zeros(shape)
    point_charges_orca[:, 0] = point_charges[:, 3]
    point_charges_orca[:, 1:4] = point_charges[:, :3]

    if hardness:
        point_charges_orca[:, 4] = hardness
    np.savetxt(
        pc_fn,
        point_charges_orca,
        fmt="%16.10f",
        header=str(len(point_charges)),
        comments="",
    )


def parse_orca_gbw(gbw_fn):
    """Adapted from
    https://orcaforum.kofo.mpg.de/viewtopic.php?f=8&t=3299

    The first 5 long int values represent pointers into the file:

    Pointer @+0:  Internal ORCA data structures
    Pointer @+8:  Geometry
    Pointer @+16: BasisSet
    Pointer @+24: Orbitals
    Pointer @+32: ECP data
    """

    with open(gbw_fn, "rb") as handle:
        handle.seek(24)
        offset = struct.unpack("<q", handle.read(8))[0]
        handle.seek(offset)
        operators = struct.unpack("<i", handle.read(4))[0]
        dimension = struct.unpack("<i", handle.read(4))[0]

        coeffs_fmt = "<" + dimension**2 * "d"
        assert operators == 1, "Unrestricted case is not implemented!"

        for i in range(operators):
            # print('\nOperator: {}'.format(i))
            coeffs = struct.unpack(coeffs_fmt, handle.read(8 * dimension**2))
            occupations = struct.iter_unpack("<d", handle.read(8 * dimension))
            energies = struct.iter_unpack("<d", handle.read(8 * dimension))
            irreps = struct.iter_unpack("<i", handle.read(4 * dimension))
            cores = struct.iter_unpack("<i", handle.read(4 * dimension))

            coeffs = np.array(coeffs).reshape(-1, dimension)
            energies = np.array([en[0] for en in energies])

        # MOs are returned in columns
        return coeffs, energies


def parse_orca_cis(cis_fn):
    """
    Read binary CI vector file from ORCA.
        Loosly based on TheoDORE 1.7.1, Authors: S. Mai, F. Plasser
        https://sourceforge.net/p/theodore-qc
    """
    cis_handle = open(cis_fn, "rb")
    # self.log(f"Parsing CI vectors from {cis_handle}")

    # The header consists of 9 4-byte integers, the first 5 of which give useful info.
    nvec = struct.unpack("i", cis_handle.read(4))[0]
    # [0] index of first alpha occ,  is equal to number of frozen alphas
    # [1] index of last  alpha occ
    # [2] index of first alpha virt
    # [3] index of last  alpha virt, header[3]+1 is equal to number of bfs
    # [4] index of first beta  occ,  for restricted equal to -1
    # [5] index of last  beta  occ,  for restricted equal to -1
    # [6] index of first beta  virt, for restricted equal to -1
    # [7] index of last  beta  virt, for restricted equal to -1
    header = [struct.unpack("i", cis_handle.read(4))[0] for i in range(8)]

    def parse_header(header):
        first_occ, last_occ, first_virt, last_virt = header
        frozen = first_occ
        occupied = last_occ + 1
        active = occupied - frozen
        mo_num = last_virt + 1
        virtual = mo_num - first_virt
        return frozen, active, occupied, virtual

    a_frozen, a_active, a_occupied, a_virtual = parse_header(header[:4])
    b_header = parse_header(header[4:])
    unrestricted = all([bh != -1 for bh in (b_header)])
    b_frozen, b_active, b_occupied, b_virtual = b_header
    a_lenci = a_active * a_virtual
    b_lenci = b_active * b_virtual
    a_nel = a_frozen + a_active
    b_nel = b_frozen + b_active
    if not unrestricted:
        b_nel = a_nel
    # expect_mult = a_nel - b_nel + 1

    # Loop over states. For non-TDA order is: X+Y of 1, X-Y of 1,
    # X+Y of 2, X-Y of 2, ...
    prev_root = -1
    prev_mult = None
    iroot_triplets = 0

    # Flags that may later be set to True
    triplets = False
    spin_flip = False
    tda = False
    Xs_a = list()
    Ys_a = list()
    Xs_b = list()
    Ys_b = list()

    def parse_coeffs(lenci, frozen, occupied, virtual):
        coeffs = struct.unpack(lenci * "d", cis_handle.read(lenci * 8))
        coeffs = np.array(coeffs).reshape(-1, virtual)
        # create full array, i.e nocc x nvirt
        coeffs_full = np.zeros((occupied, virtual))
        coeffs_full[frozen:] = coeffs
        return coeffs_full

    def handle_X_Y(root_updated, Xs, Ys, coeffs):
        # When the root was not incremented compared to the previous root we have
        # just parsed X-Y (and parsed X+Y before.)
        #
        # We can recover the separate X and Y vectors by first computing X as
        #   X = (X + Y + X - Y) / 2
        # and then
        #   Y = X + Y - X
        if root_updated:
            X_plus_Y = Xs[-1]  # Parsed in previous cycle
            X_minus_Y = coeffs  # Parsed in current cycle
            X = 0.5 * (X_plus_Y + X_minus_Y)
            Y = X_plus_Y - X
            # Update the X and Y vectors that were already saved with their correct values.
            Xs[-1] = X
            Ys[-1] = Y
        # When the root was incremented we either have a TDA calculation without Y or
        # we parsed X-Y in the previous cycle and now moved to a new root.
        else:
            Xs.append(coeffs)
            Ys.append(np.zeros_like(coeffs))

    for ivec in range(nvec):
        # Header of each vector, contains 6 4-byte ints.
        ncoeffs, _, mult, _, iroot, _ = struct.unpack("iiiiii", cis_handle.read(24))

        # Check if we deal with a spin-flip calculation. There, excitations are from
        # α_activate -> β_virtual.
        if unrestricted and ncoeffs == (a_active * b_virtual):
            unrestricted = False  # Don't expect β_active -> β_virtual.
            spin_flip = True
            a_lenci = ncoeffs
            a_virtual = b_virtual
            warnings.warn(
                "Spin-flip calculation detected. Pysisyphus can parse it, but "
                "the transition density matrix is not yet handled properly by the "
                "OverlapCalculator-class or the Wavefunction-class!"
            )

        if prev_mult is None:
            prev_mult = mult

        # 2 x 8 bytes unknown?!
        ene, _ = struct.unpack("dd", cis_handle.read(16))

        # Will evaluate True only once when triplets were requested.
        if prev_mult != mult:
            triplets = True
            prev_root = -1

        # When we encounter the second "state" we can decide if it is a TDA
        # calculation (without Y-vector).
        if (ivec == 1) and (iroot == prev_root + 1):
            tda = True

        if triplets:
            iroot = iroot_triplets

        root_updated = prev_root == iroot

        # self.log(f"{ivec=}, {nele=}, {mult=}, {iroot=}, {root_updated=}")
        # Then come nact * nvirt 8-byte doubles with the coefficients
        coeffs_a = parse_coeffs(a_lenci, a_frozen, a_occupied, a_virtual)
        handle_X_Y(root_updated, Xs_a, Ys_a, coeffs_a)
        if unrestricted:
            coeffs_b = parse_coeffs(b_lenci, b_frozen, b_occupied, b_virtual)
            handle_X_Y(root_updated, Xs_b, Ys_b, coeffs_b)

        # Somehow ORCA stops to update iroot correctly after the singlet states.
        if (mult == 3) and (tda or (ivec % 2) == 1):
            iroot_triplets += 1

        prev_root = iroot
        prev_mult = mult
    # Verify, that we are at the EOF. We request 1 byte, but we only get 0.
    assert len(cis_handle.read(1)) == 0
    cis_handle.close()

    # Convert everything to numpy arrays.
    Xs_a, Ys_a, Xs_b, Ys_b = [np.array(_) for _ in (Xs_a, Ys_a, Xs_b, Ys_b)]

    def handle_triplets(Xs, Ys):
        assert (len(Xs) % 2) == 0
        states = len(Xs) // 2
        Xs = Xs[states:]
        Ys = Ys[states:]
        return Xs, Ys

    # Only return triplet states if present
    if triplets:
        Xs_a, Ys_a = handle_triplets(Xs_a, Ys_a)
        assert len(Xs_b) == 0
        assert len(Ys_b) == 0

    # Beta part will be empty
    if not unrestricted:
        assert len(Xs_b) == 0
        assert len(Ys_b) == 0
        Xs_b = np.zeros_like(Xs_a)
        Ys_b = np.zeros_like(Xs_b)

    return Xs_a, Ys_a, Xs_b, Ys_b


@file_or_str(".log", ".out")
def parse_orca_all_energies(text, triplets=False, do_tddft=False):
    energy_re = r"FINAL SINGLE POINT ENERGY\s*([-\.\d]+)"
    energy_mobj = re.search(energy_re, text)
    gs_energy = float(energy_mobj.groups()[0])
    all_energies = [gs_energy]

    if do_tddft:
        scf_re = re.compile(r"E\(SCF\)\s+=\s*([\d\-\.]+) Eh")
        scf_mobj = scf_re.search(text)
        scf_en = float(scf_mobj.group(1))
        gs_energy = scf_en
        tddft_re = re.compile(r"STATE\s*(\d+):\s*E=\s*([\d\.]+)\s*au")
        states, exc_ens = zip(
            *[(int(state), float(en)) for state, en in tddft_re.findall(text)]
        )
        if triplets:
            roots = len(states) // 2
            exc_ens = exc_ens[-roots:]
            states = states[-roots:]
        assert len(exc_ens) == len(set(states))
        all_energies = np.full(1 + len(exc_ens), gs_energy)
        all_energies[1:] += exc_ens
    all_energies = np.array(all_energies)
    return all_energies


def get_name(text: bytes):
    """Return string that comes before first \x00 character & offset."""
    until = text.find(b"\x00")
    return text[:until].decode(), until


@file_or_str(".densities", mode="rb")
def parse_orca_densities(text: bytes):
    handle = io.BytesIO(text)

    # Determine file size
    handle.seek(0, 2)
    file_size = handle.tell()
    handle.seek(0, 0)

    offset, _ = struct.unpack(
        "ii", handle.read(8)
    )  # Don't know about the second integer
    # print("offset", offset)
    dens_size = offset - 8
    assert dens_size % 8 == 0
    dens_floats = dens_size // 8
    # print(f"Expecting {dens_floats} density doubles")
    densities = struct.unpack("d" * dens_floats, handle.read(dens_size))
    ndens = struct.unpack("i", handle.read(4))[0]

    # Block of 512 bytes meta data. I don't really know what is contained in there.
    meta = handle.read(512)
    base_name, _ = get_name(meta)

    # Now multiple 512 byte blocks for each density follow
    dens_names = list()
    for i in range(ndens):
        dens_meta = handle.read(512)
        dens_name, _ = get_name(dens_meta)
        dens_names.append(dens_name)
        # don't know about the first item, 2nd items seems to 0
        _, _, nao1, nao2 = struct.unpack("iiii", handle.read(16))
        assert nao1 == nao2
        _ = struct.unpack("b", handle.read(1))[0]  # 0 byte
        assert _ == 0
    dens_exts = [Path(dens_name).suffix[1:] for dens_name in dens_names]

    # Verify that we parsed to whole file
    assert file_size - handle.tell() == 0
    handle.close()

    # Construct density matrices
    assert dens_floats % ndens == 0
    nao = int(sqrt(dens_floats // ndens))
    dens_shape = (nao, nao)
    densities = np.array(densities).reshape(ndens, *dens_shape)

    dens_dict = {dens_ext: dens for dens_ext, dens in zip(dens_exts, densities)}
    # This check could be removed but I'll keep if for now, so I only deal with
    # known densities.
    # scfp : HF/DFT electronic density
    # scfr : HF/DFT spin density
    # cisp : TDA/TD-DFT/CIS electronic density
    # cisr : TDA/TD-DFT/CIS spin density
    assert set(dens_dict) <= {"scfp", "scfr", "cisp", "cisr"}

    return dens_dict


def get_exc_ens_fosc(wf_fn, cis_fn, log_fn):
    wf = Wavefunction.from_file(wf_fn)
    Xa, Ya, Xb, Yb = parse_orca_cis(cis_fn)
    all_energies = parse_orca_all_energies(log_fn, do_tddft=True)
    Xa, Ya = norm_ci_coeffs(Xa, Ya)
    exc_ens = all_energies[1:] - all_energies[0]
    tdens = wf.get_transition_dipole_moment(Xa + Ya)
    warnings.warn("Only alpha TDM is currently taken into account!")
    fosc = wf.oscillator_strength(exc_ens, tdens)
    return exc_ens, fosc


class ORCA(OverlapCalculator):

    conf_key = "orca"
    _set_plans = (
        "gbw",
        "out",
        "cis",
        "densities",
        ("molden", "mwfn_wf"),
    )

    def __init__(
        self,
        keywords,
        blocks="",
        gbw=None,
        do_stable=False,
        numfreq=False,
        json_dump=True,
        **kwargs,
    ):
        """ORCA calculator.

        Wrapper for creating ORCA input files for energy, gradient
        and Hessian calculations. The PAL and memory inputs must not
        be given in the keywords and/or blocks, as they are handled
        by the 'pal' and 'memory' arguments.

        Parameters
        ----------
        keywords : str
            Keyword line, as normally given in ORCA, excluding the
            leading "!".
        blocks : str, optional
            ORCA block input(s), e.g. for TD-DFT calculations (%tddft ... end).
            As the blocks start with a leading "%", wrapping the input in quotes
            ("") is required, otherwise the parsing will fail.
        gbw : str, optional
            Path to an input gbw file, which will be used as initial guess
            for the first calculation. Will be overriden later, with the
            path to the gbw file of a previous calculation.
        do_stable: bool, optional
            Run stability analysis until a stable wavefunction is obtained,
            before every calculation.
        numfreq : bool, optional
            Use numerical frequencies instead of analytical ones.
        json_dump : bool, optional
            Whether to dump the wavefunction to JSON via orca_2json. The JSON can become
            very large in calculations comprising many basis functions.
        """
        super().__init__(**kwargs)

        self.keywords = keywords.lower()
        self.blocks = blocks.lower()
        self.gbw = gbw
        self.do_stable = bool(do_stable)
        self.freq_keyword = "numfreq" if numfreq else "freq"
        self.json_dump = bool(json_dump)

        assert ("pal" not in keywords) and ("nprocs" not in blocks), (
            "PALn/nprocs not " "allowed! Use 'pal: n' in the 'calc' section instead."
        )
        assert "maxcore" not in blocks, (
            "maxcore not allowed! " "Use 'mem: n' in the 'calc' section instead!"
        )

        self.to_keep = (
            "inp",
            "out:orca.out",
            "gbw",
            "engrad",
            "hessian",
            "cis",
            "molden:orca.molden",
            "hess",
            "pcgrad",
            "densities:orca.densities",
            "json",
        )
        self.do_tddft = False
        if "tddft" in self.blocks:
            self.do_tddft = True
            try:
                self.root = int(re.search(r"iroot\s*(\d+)", self.blocks).group(1))
            except AttributeError:
                self.log("Doing TDA/TDDFT calculation without gradient.")
        self.triplets = bool(re.search(r"triplets\s+true", self.blocks))
        self.inp_fn = "orca.inp"
        self.out_fn = "orca.out"

        self.orca_input = """!{keywords} {calc_type}
        {moinp}

        %pal nprocs {pal} end
        %maxcore {mem}

        {blocks}
        {pointcharges}

        *xyz {charge} {mult}
        {coords}
        *
        """

        self.parser_funcs = {
            "energy": self.parse_energy,
            "grad": self.parse_engrad,
            "hessian": self.parse_hessian,
            "noparse": lambda path: None,
            "stable": self.parse_stable,
        }

        self.base_cmd = self.get_cmd()

    def reattach(self, last_calc_cycle):
        # Use the latest .gbw
        gbw = self.make_fn("gbw", last_calc_cycle)
        self.log(f"Restarted. using {gbw}")

    def get_moinp_str(self, gbw):
        moinp_str = ""
        if gbw:
            moinp_str = f"""!moread
            %moinp "{os.path.abspath(gbw)}" """
        return moinp_str

    def prepare_input(
        self, atoms, coords, calc_type, point_charges=None, do_stable=False
    ):
        coords = self.prepare_coords(atoms, coords)
        if self.gbw:
            self.log(f"Using {self.gbw}")
        else:
            self.log("Using initial guess provided by ORCA")
        if calc_type == "noparse":
            calc_type = ""

        pc_str = ""
        if point_charges is not None:
            pc_fn = self.make_fn("pointcharges_inp.pc")
            save_orca_pc_file(point_charges, pc_fn)
            pc_str = f'%pointcharges "{pc_fn}"'
        stable_block = "\n%scf stabperform true hftyp uhf end" if do_stable else ""
        blocks = self.get_block_str() + stable_block

        inp = self.orca_input.format(
            keywords=self.keywords,
            calc_type=calc_type,
            moinp=self.get_moinp_str(self.gbw),
            pal=self.pal,
            mem=self.mem,
            blocks=blocks,
            pointcharges=pc_str,
            coords=coords,
            charge=self.charge,
            mult=self.mult,
        )
        return inp

    def get_block_str(self):
        block_str = self.blocks
        # Use the correct root if we track it
        if self.track:
            block_str = re.sub(r"iroot\s+(\d+)", f"iroot {self.root}", self.blocks)
            self.log(f"Using iroot '{self.root}' for excited state gradient.")
        return block_str

    def get_stable_wavefunction(self, atoms, coords):
        self.log("Trying to get a stable wavefunction")

        stable = False
        max_cycles = 10
        for i in range(max_cycles):
            inp = self.prepare_input(atoms, coords, calc_type="", do_stable=True)
            stable = self.run(inp, calc="stable")
            self.log(f"{i:02d} stable: {stable}")

            if stable:
                self.log(f"Found stable wavefunction in cycle {i}!")
                break
        else:
            raise Exception(
                "Could not find stable wavefunction in {max_cycles}! " "Aborting."
            )

    def parse_stable(self, path):
        with open(path / "orca.out") as handle:
            text = handle.read()

        stable_re = re.compile("Stability Analysis indicates a stable")
        stable = bool(stable_re.search(text))

        unstable_re = re.compile("Stability Analysis indicates an UNSTABLE")
        unstable = bool(unstable_re.search(text))

        stable = stable and not unstable

        return stable

    def store_and_track(self, results, func, atoms, coords, **prepare_kwargs):
        if self.track:
            self.store_overlap_data(atoms, coords)
            if self.track_root():
                # Redo the calculation with the updated root
                results = func(atoms, coords, **prepare_kwargs)
        results["all_energies"] = self.parse_all_energies()
        return results

    def get_energy(self, atoms, coords, **prepare_kwargs):
        calc_type = ""

        if self.do_stable:
            self.get_stable_wavefunction(atoms, coords)

        inp = self.prepare_input(atoms, coords, calc_type, **prepare_kwargs)
        results = self.run(inp, calc="energy")
        results = self.store_and_track(
            results, self.get_energy, atoms, coords, **prepare_kwargs
        )
        return results

    def get_forces(self, atoms, coords, **prepare_kwargs):
        if self.do_stable:
            self.get_stable_wavefunction(atoms, coords)

        calc_type = "engrad"
        inp = self.prepare_input(atoms, coords, calc_type, **prepare_kwargs)
        kwargs = {
            "calc": "grad",
        }
        results = self.run(inp, **kwargs)
        results = self.store_and_track(
            results, self.get_forces, atoms, coords, **prepare_kwargs
        )
        return results

    def get_hessian(self, atoms, coords, **prepare_kwargs):
        calc_type = self.freq_keyword

        if self.do_stable:
            self.get_stable_wavefunction(atoms, coords)

        inp = self.prepare_input(atoms, coords, calc_type, **prepare_kwargs)
        results = self.run(inp, calc="hessian")
        # results = self.store_and_track(
        # results, self.get_hessian, atoms, coords, **prepare_kwargs
        # )
        return results

    def run_calculation(self, atoms, coords, **prepare_kwargs):
        """Basically some kind of dummy method that can be called
        to execute ORCA with the stored cmd of this calculator."""
        inp = self.prepare_input(atoms, coords, "noparse", **prepare_kwargs)
        kwargs = {
            "calc": "energy",
        }
        results = self.run(inp, **kwargs)
        if self.track:
            self.store_overlap_data(atoms, coords)
        return results

    def run_after(self, path):
        # Create .molden file when CDDs are requested
        if self.cdds:
            cmd = "orca_2mkl orca -molden"
            self.popen(cmd, cwd=path)
            shutil.copy(path / "orca.molden.input", path / "orca.molden")

        if self.json_dump: 
            # Will silently fail with ECPs
            cmd = "orca_2json orca"
            proc = self.popen(cmd, cwd=path)
            if (ret := proc.returncode) != 0:
                self.log(f"orca_2json call failed with return-code {ret}!")

    @staticmethod
    @file_or_str(".hess", method=False)
    def parse_hess_file(text):
        integer = pp.Word(pp.nums)
        float_ = pp.Word(pp.nums + ".-")
        plus = pp.Literal("+")
        minus = pp.Literal("-")
        E = pp.Literal("E")
        scientific = pp.Combine(float_ + E + pp.Or([plus, minus]) + integer)

        table_header_line = pp.Suppress(integer + pp.restOfLine)
        scientific_line = pp.Suppress(integer) + pp.OneOrMore(scientific)
        scientific_block = table_header_line + pp.OneOrMore(scientific_line)
        float_line = pp.Suppress(integer) + float_
        comment_line = pp.Literal("#") + pp.restOfLine
        mass_xyz_line = pp.Group(
            pp.Word(pp.alphas) + float_ + pp.Group(pp.OneOrMore(float_))
        )

        block_name = pp.Word(pp.alphas + "$_")
        block_length = integer

        block_int = block_name + block_length
        block_float = block_name + float_
        block_table = block_name + integer + pp.OneOrMore(scientific_block)
        block_table_two_int = (
            block_name + integer + pp.Suppress(integer) + pp.OneOrMore(scientific_block)
        )
        block_float_table = block_name + integer + pp.OneOrMore(float_line)
        block_atoms = block_name + integer + pp.OneOrMore(mass_xyz_line)

        act_atom = block_int.setResultsName("act_atom")
        act_coord = block_int.setResultsName("act_coord")
        act_energy = block_float.setResultsName("act_energy")
        hessian = block_table.setResultsName("hessian")
        vib_freqs = block_float_table.setResultsName("vib_freqs")
        normal_modes = block_table_two_int.setResultsName("normal_modes")
        atoms = block_atoms.setResultsName("atoms")

        parser = (
            block_name
            + act_atom
            + act_coord
            + act_energy
            + hessian
            + vib_freqs
            + normal_modes
            + pp.OneOrMore(comment_line)
            + atoms
        )
        parsed = parser.parseString(text)
        return parsed

    def parse_hessian(self, path):
        hessian_fn = glob.glob(os.path.join(path, "*.hess"))
        assert len(hessian_fn) == 1
        hessian_fn = hessian_fn[0]
        if not hessian_fn:
            raise Exception("ORCA calculation failed.")

        parsed = ORCA.parse_hess_file(hessian_fn)

        # logging.warning("Hacky orca energy parsing in orca hessian calculation!")
        orca_log_fn = os.path.join(path, self.out_fn)
        with open(orca_log_fn) as handle:
            log_text = handle.read()

        energy_re = r"FINAL SINGLE POINT ENERGY\s*([-\.\d]+)"
        energy_mobj = re.search(energy_re, log_text)
        energy = float(energy_mobj.groups()[0])

        results = {
            "energy": energy,
            "hessian": make_sym_mat(parsed["hessian"]),
        }

        return results

    def parse_energy(self, path):
        log_fn = glob.glob(os.path.join(path / "orca.out"))
        if not log_fn:
            raise Exception("ORCA calculation failed.")

        assert len(log_fn) == 1
        log_fn = log_fn[0]
        with open(log_fn) as handle:
            text = handle.read()
        mobj = re.search(r"FINAL SINGLE POINT ENERGY\s+([\d\-\.]+)", text)
        energy = float(mobj[1])
        return {"energy": energy}

    def parse_engrad(self, path):
        results = {}
        engrad_fn = glob.glob(os.path.join(path, "*.engrad"))
        if not engrad_fn:
            raise Exception("ORCA calculation failed.")

        assert len(engrad_fn) == 1
        engrad_fn = engrad_fn[0]
        with open(engrad_fn) as handle:
            engrad = handle.read()
        engrad = re.findall(r"([\d\-\.]+)", engrad)
        atoms = int(engrad.pop(0))
        energy = float(engrad.pop(0))
        force = -np.array(engrad[: 3 * atoms], dtype=float)
        results["energy"] = energy
        results["forces"] = force

        return results

    @staticmethod
    def parse_cis(cis):
        """Simple wrapper of external function.

        Currently, only returns Xα and Yα.
        """
        return parse_orca_cis(cis)[:2]

    @staticmethod
    def parse_gbw(gbw_fn):
        return parse_orca_gbw(gbw_fn)

    @staticmethod
    def set_mo_coeffs_in_gbw(in_gbw_fn, out_gbw_fn, mo_coeffs):
        """See self.parse_gbw."""

        with open(in_gbw_fn, "rb") as handle:
            handle.seek(24)
            offset = struct.unpack("<q", handle.read(8))[0]
            handle.seek(offset)
            operators = struct.unpack("<i", handle.read(4))[0]
            dimension = struct.unpack("<i", handle.read(4))[0]
            assert operators == 1, "Unrestricted case is not implemented!"

            handle.seek(0)
            gbw_bytes = handle.read()

        tot_offset = offset + 4 + 4
        start = gbw_bytes[:tot_offset]
        end = gbw_bytes[tot_offset + 8 * dimension**2 :]
        # Construct new gbw content by replacing the MO coefficients in the middle
        mod_gbw_bytes = start + mo_coeffs.T.tobytes() + end

        with open(out_gbw_fn, "wb") as handle:
            handle.write(mod_gbw_bytes)

    def parse_all_energies(self, text=None, triplets=None):
        if text is None:
            with open(self.out) as handle:
                text = handle.read()

        if triplets is None:
            triplets = self.triplets

        return parse_orca_all_energies(text, triplets, self.do_tddft)

    @staticmethod
    @file_or_str(".out", method=False)
    def parse_atoms_coords(text):
        ac_re = re.compile(
            r"CARTESIAN COORDINATES \(ANGSTROEM\)\s+\-{33}(.+?)\s+\-{28}", re.DOTALL
        )
        mobj = ac_re.search(text)
        atoms_coords = mobj.group(1).strip().split()
        # atoms, *coords = np.array(atoms_coords).reshape(-1, 4).T
        atoms_coords = np.array(atoms_coords).reshape(-1, 4)
        atoms = tuple(atoms_coords[:, 0])
        coords = atoms_coords[:, 1:].astype(float).flatten() * ANG2BOHR
        return atoms, coords

    @staticmethod
    @file_or_str(".out", method=False)
    def parse_engrad_info(text):
        soi_re = re.compile(r"State of interest\s+\.{3}\s+(\d+)")
        try:
            root = soi_re.search(text).group(1)
            root = int(root)
        except AttributeError:
            root = None
        triplets = bool(re.search(r"triplets\s+true", text))
        return root, triplets

    def parse_mo_numbers(self, out_fn):
        with open(out_fn) as handle:
            text = handle.read()
        electron_re = r"NEL\s*....\s*(\d+)"
        electrons = int(re.search(electron_re, text)[1])
        assert electrons % 2 == 0, "unrestricted is not yet supported!"
        occ_num = int(electrons / 2)

        mo_re = r"Dim\s*....\s*(\d+)"
        mo_num = int(re.search(mo_re, text)[1])
        virt_num = mo_num - occ_num
        self.log(
            f"found {electrons} electrons, {mo_num} MOs, with "
            f"{occ_num} occupied and {virt_num} virtual."
        )
        return occ_num, virt_num

    def set_mo_coeffs(self, mo_coeffs=None, gbw=None):
        if mo_coeffs is not None:
            self.mo_coeffs = mo_coeffs
            return
        if not gbw and self.gbw:
            gbw = self.gbw
        else:
            raise Exception("Got no .gbw file to parse!")
        self.log(f"Setting MO coefficients from {gbw}.")
        self.mo_coeffs, _ = self.parse_gbw(self.gbw)

    def prepare_overlap_data(self, path):
        # Parse eigenvectors from tda/tddft calculation
        X, Y = self.parse_cis(self.cis)
        # Parse mo coefficients from gbw file and write a 'fake' turbomole
        # mos file.
        C, _ = self.parse_gbw(self.gbw)
        all_energies = self.parse_all_energies()
        return C, X, Y, all_energies

    def get_chkfiles(self):
        return {
            "gbw": self.gbw,
        }

    def set_chkfiles(self, chkfiles):
        try:
            gbw = chkfiles["gbw"]
            self.gbw = gbw
            self.log(f"Set chkfile '{gbw}' as gbw.")
        except KeyError:
            self.log("Found no gbw information in chkfiles!")

    @file_or_str(".out", method=True)
    def check_termination(self, text):
        term_re = re.compile(r"\*{4}ORCA TERMINATED NORMALLY\*{4}")
        mobj = term_re.search(text)
        return bool(mobj)

    def clean_tmp(self, path):
        tmp_fns = path.glob("*.tmp")
        for tmp in tmp_fns:
            os.remove(tmp)
            self.log(f"Removed '{tmp}'")
        # try:
        # os.remove(path / "orca.gbw")
        # except FileNotFoundError:
        # pass

    def __str__(self):
        return f"ORCA({self.name})"



import glob
import os

import numpy as np
import pyparsing as pp
import re


def to_float(s, loc, toks):
    match = toks[0].replace("D", "E")
    return float(match)


def make_float_class(**kwargs):
    return pp.Word(pp.nums + ".-DE+", **kwargs).setParseAction(to_float)


def parse_turbo_gradient(path):
    results = {}
    gradient_fn = glob.glob(os.path.join(path, "gradient"))
    if not gradient_fn:
        raise Exception("gradient file not found!")
    assert len(gradient_fn) == 1
    gradient_fn = gradient_fn[0]
    with open(gradient_fn) as handle:
        text = handle.read()

    float_ = make_float_class()
    cycle = pp.Word(pp.nums).setResultsName("cycle")
    scf_energy = float_.setResultsName("scf_energy")
    grad_norm = float_.setResultsName("grad_norm")
    float_line = float_ + float_ + float_
    coord_line = pp.Group(float_line + pp.Word(pp.alphas))
    grad_line = pp.Group(float_line)
    cart_grads = pp.Literal("cartesian gradients")
    energy_type = pp.Or(
        (
            pp.Literal("SCF energy"),
            pp.Literal("ex. state energy"),
            pp.Literal("CC2 energy"),
            pp.Literal("ADC(2) energy"),
            pp.Literal("MP2 energy"),
        )
    )

    parser = (
        pp.Or((pp.Literal("$grad"), pp.Literal("$gradient")))
        + pp.Optional(cart_grads)
        + pp.Literal("cycle =")
        + cycle
        + energy_type
        + pp.Literal("=")
        + scf_energy
        + pp.Literal("|dE/dxyz| =")
        + grad_norm
        + pp.OneOrMore(coord_line).setResultsName("coords")
        + pp.OneOrMore(grad_line).setResultsName("grad")
        + pp.Literal("$end")
    )
    parsed = parser.parseString(text)
    gradient = np.array(parsed["grad"].asList()).flatten()

    results["energy"] = parsed["scf_energy"]
    results["forces"] = -gradient
    return results


def parse_turbo_ccre0_ascii(text):
    float_ = make_float_class()
    float_20 = make_float_class(exact=20)
    int_ = pp.Word(pp.nums).setParseAction(lambda s, loc, toks: int(toks[0]))

    title = pp.Literal("$CCRE0-") + pp.Word(pp.nums + "-")
    method = pp.Word(pp.alphanums + "()") + float_ + float_ + int_ + int_ + int_
    data = pp.Group(pp.OneOrMore(float_20))
    end = pp.Literal("$end")

    parser = title + method + data.setResultsName("data") + end
    result = parser.parseString(text)
    data = np.array(result["data"].asList())
    return data


def parse_turbo_mos(text):
    float_20 = make_float_class(exact=20)
    int_ = pp.Word(pp.nums)
    comment = pp.Literal("#") + pp.restOfLine

    mo_num = int_
    sym = pp.Word(pp.alphanums)
    eigenvalue = pp.Literal("eigenvalue=") + float_20
    nsaos = pp.Literal("nsaos=") + int_
    mo_coeffs = pp.OneOrMore(float_20)

    mo = pp.Group(
        mo_num + sym + eigenvalue + nsaos + mo_coeffs.setResultsName("mo_coeffs")
    )

    parser = (
        pp.Literal("$scfmo")
        + pp.Literal("scfconv=")
        + pp.Word(pp.nums)
        + pp.Literal("format(4d20.14) ")
        + pp.ZeroOrMore(comment)
        + pp.OneOrMore(mo).setResultsName("mos")
        + pp.Literal("$end")
    )
    parsed = parser.parseString(text)
    mo_coeffs = np.array([mo.mo_coeffs.asList() for mo in parsed.mos]).T

    # MOs are in columns
    return mo_coeffs


def parse_turbo_exstates(text):
    """Parse excitation energies (first blocks) from an exstates file."""
    float_ = make_float_class()
    exc_energies_line = (
        pp.Literal("$excitation_energies_")
        + pp.Word(pp.alphanums + "()").setResultsName("model")
        + pp.Word("_")
        + pp.restOfLine
    )
    exc_energy = pp.Suppress(pp.Word(pp.nums)) + float_
    exc_energies_block = pp.Group(
        exc_energies_line + pp.Group(pp.OneOrMore(exc_energy)).setResultsName("exc_ens")
    )

    parser = pp.OneOrMore(exc_energies_block).setResultsName("exc_blocks")
    result = parser.parseString(text)
    exc_energies_by_model = [(b.model, b.exc_ens.asList()) for b in result.exc_blocks]
    return exc_energies_by_model


def parse_turbo_exstates_re(text):
    results = dict()
    exc_ens_str = "excitation_energies_"
    model_re = re.compile(exc_ens_str + r"(\w+?)_")
    blocks = text.strip().split("$")
    for block in blocks:
        if not block.startswith(exc_ens_str):
            continue
        model_line, *exc_lines = block.strip().split("\n")
        mobj = model_re.search(model_line)
        model = mobj.group(1)
        exc_lines = [line.strip().split() for line in exc_lines]
        assert all([len(line) == 2 for line in exc_lines])
        exc_ens = np.array([exc_en for _, exc_en in exc_lines], dtype=float)
        results[model] = exc_ens
    return results



import numpy as np
import sympy as sym
from sympy import atan, exp, tan, sin, pi


from pysisyphus.calculators.AnaPotBase import AnaPotBase
from pysisyphus.calculators.Calculator import Calculator

# https://www.wolframalpha.com/input/?i=plot+arccot(-exp(y)*cot(x/2-pi/4))+-+2*exp(-(y-sin(x))^2/2)
# [1] http://aip.scitation.org/doi/abs/10.1063/1.461606
# https://www.wolframalpha.com/input/?i=derivative+of+(arccot(-exp(y)*cot(x/2-pi/4))+-+2*exp(-(y-sin(x))^2/2))

class AnaPot2_(Calculator):

    def __init__(self): 
        super().__init__()

    def get_energy(self, atoms, coords):
        x, y, z = coords
        cot = 1 / np.tan(x/2 - np.pi/4)
        arccot = np.arctan(
                1 / (-np.exp(y) * cot)
        )
        energy = (
            arccot - 2 * np.exp(-(y-np.sin(x))**2 / 2)
        )
        return {"energy": energy}

    def __str__(self):
        return "AnaPot2 calculator"


class AnaPot2(AnaPotBase):
    """We can't use sympify as it replaces 1/tan by cot and this isn't
    supported by numpy when we call lambdify."""

    def __init__(self): 
        x, y = sym.symbols("x y")
        V_str = atan(exp(-y)/tan(x/2 + pi/4)) - 2*exp(-(y - sin(x))**2/2)
        # xlim = (-np.pi/2, np.pi)
        xlim = (-np.pi, np.pi)
        ylim = (-2, 2)
        levels = np.linspace(-2, 1, 40)
        super().__init__(V_str=V_str, xlim=xlim, ylim=ylim, levels=levels,
                         use_sympify=False)

    def __str__(self):
        return "AnaPot2 calculator"



from pysisyphus.constants import BOHR2ANG

class FakeASE:

    def __init__(self, calc):
        self.calc = calc

        self.results = dict()

    def get_atoms_coords(self, atoms):
        return (atoms.get_chemical_symbols(),
                # Convert ASE Angstrom to Bohr for pysisyphus
                atoms.get_positions().flatten() / BOHR2ANG
        )

    def get_potential_energy(self, atoms=None):
        atoms, coords = self.get_atoms_coords(atoms)
        results = self.calc.get_energy(atoms, coords)

        energy = results["energy"]
        # self.results["energy"] = results["energy"]

        return energy

    def get_forces(self, atoms=None):
        atoms, coords = self.get_atoms_coords(atoms)
        results = self.calc.get_forces(atoms, coords)

        # Convert back to Angstrom for ASE
        forces = results["forces"].reshape(-1, 3) / BOHR2ANG

        # self.results["energy"] = results["energy"]
        # self.results["forces"] = forces

        return forces



__all__ = [
    "geom_from_cjson",
    "geom_from_crd",
    "geom_from_hessian",
    "geom_from_mol2",
    "geom_from_pdb",
    "geom_from_zmat",
    "geom_from_qcschema",
    "geoms_from_xyz",
    "save_hessian",
    "save_third_deriv",
]


from pysisyphus.io.cjson import geom_from_cjson
from pysisyphus.io.crd import geom_from_crd, geom_to_crd_str
from pysisyphus.io.hessian import save_hessian, save_third_deriv, geom_from_hessian
from pysisyphus.io.mol2 import geom_from_mol2
from pysisyphus.io.pdb import geom_from_pdb
from pysisyphus.io.qcschema import geom_from_qcschema
from pysisyphus.io.xyz import geoms_from_xyz, geoms_from_inline_xyz, parse_xyz
from pysisyphus.io.zmat import geom_from_zmat, geom_from_zmat_fn



import logging
import sys

logger = logging.getLogger("pysisyphus")
logger.setLevel(logging.DEBUG)

file_handler = logging.FileHandler("pysisyphus.log", mode="w", delay=True)
logger.addHandler(file_handler)

stdout_handler = logging.StreamHandler(sys.stdout)
stdout_handler.setLevel(logging.INFO)
logger.addHandler(stdout_handler)



"""Unit tests for shared DefaultGroup behavior."""

from __future__ import annotations

import click
from click.testing import CliRunner

from pdb2reaction.advanced_help import (
    _configure_subcommand_help_visibility,
    _ensure_help_advanced_option,
)
from pdb2reaction.bool_compat import normalize_bool_argv as _normalize_bool_argv_impl
from pdb2reaction.default_group import DefaultGroup


def _normalize_passthrough(args, *_):
    return args, False


def _normalize_with_legacy_flag(args, *_):
    return args, True


def _make_group(
    *,
    default: str | None = None,
    lazy_subcommands: dict[str, tuple[str, str, str]] | None = None,
    normalize_bool_argv=_normalize_passthrough,
    primary_help_options: dict[str, frozenset[str]] | None = None,
    parser_wrapper_subcommands: frozenset[str] | None = None,
    parser_wrapper_bool_option_providers=None,
) -> click.Group:
    @click.group(
        cls=DefaultGroup,
        default=default,
        lazy_subcommands=lazy_subcommands or {},
        command_bool_value_options={},
        command_bool_toggle_options={},
        command_bool_toggle_negative_aliases={},
        parser_wrapper_subcommands=parser_wrapper_subcommands or frozenset(),
        parser_wrapper_bool_option_providers=parser_wrapper_bool_option_providers or {},
        subcommand_primary_help_options=primary_help_options or {},
        normalize_bool_argv=normalize_bool_argv,
        ensure_help_advanced_option=_ensure_help_advanced_option,
        configure_subcommand_help_visibility=_configure_subcommand_help_visibility,
    )
    def cli() -> None:
        """Test CLI."""

    return cli


def test_default_subcommand_is_inserted_when_no_args() -> None:
    cli = _make_group(default="all")

    @cli.command(name="all")
    def all_cmd() -> None:
        click.echo("all-called")

    runner = CliRunner()
    result = runner.invoke(cli, [])
    assert result.exit_code == 0
    assert "all-called" in result.output


def test_help_does_not_trigger_default_subcommand() -> None:
    cli = _make_group(default="all")

    @cli.command(name="all")
    def all_cmd() -> None:
        click.echo("all-called")

    runner = CliRunner()
    result = runner.invoke(cli, ["--help"])
    assert result.exit_code == 0
    assert "all-called" not in result.output
    assert "Commands:" in result.output


def test_legacy_bool_syntax_does_not_emit_deprecation_message() -> None:
    cli = _make_group(normalize_bool_argv=_normalize_with_legacy_flag)

    @cli.command()
    def ping() -> None:
        click.echo("ping-called")

    runner = CliRunner()
    result = runner.invoke(cli, ["ping"])
    assert result.exit_code == 0
    assert "ping-called" in result.output
    assert "Legacy bool syntax '--flag True/False'" not in result.output


def test_help_advanced_shows_hidden_options() -> None:
    primary = {"scan": frozenset({"--input", "--help-advanced"})}
    cli = _make_group(primary_help_options=primary)

    @cli.command(name="scan")
    @click.option("--input")
    @click.option("--advanced-opt")
    def scan_cmd(input: str | None, advanced_opt: str | None) -> None:
        _ = (input, advanced_opt)
        click.echo("scan-called")

    runner = CliRunner()
    short = runner.invoke(cli, ["scan", "--help"])
    assert short.exit_code == 0
    assert "--help-advanced" in short.output
    assert "--input" in short.output
    assert "--advanced-opt" not in short.output

    advanced = runner.invoke(cli, ["scan", "--help-advanced"])
    assert advanced.exit_code == 0
    assert "--advanced-opt" in advanced.output


def test_lazy_import_failure_is_reported_as_click_exception() -> None:
    cli = _make_group(
        lazy_subcommands={
            "broken": (".__this_module_should_not_exist__", "cli", "Broken command")
        }
    )

    runner = CliRunner()
    result = runner.invoke(cli, ["broken"])
    assert result.exit_code != 0
    assert "Command 'broken' is unavailable because the module could not be imported." in result.output
    assert "Missing dependency:" in result.output


def test_bool_toggle_accepts_value_style_syntax_via_auto_detection() -> None:
    cli = _make_group(normalize_bool_argv=_normalize_bool_argv_impl)

    @cli.command(name="scan")
    @click.option("--detect-layer/--no-detect-layer", default=True)
    def scan_cmd(detect_layer: bool) -> None:
        click.echo(f"detect_layer={detect_layer}")

    runner = CliRunner()
    result_false = runner.invoke(cli, ["scan", "--detect-layer", "False"])
    assert result_false.exit_code == 0
    assert "detect_layer=False" in result_false.output

    result_no = runner.invoke(cli, ["scan", "--no-detect-layer"])
    assert result_no.exit_code == 0
    assert "detect_layer=False" in result_no.output


def test_single_flag_accepts_no_prefix_and_value_style_syntax() -> None:
    cli = _make_group(normalize_bool_argv=_normalize_bool_argv_impl)

    @cli.command(name="add-elem-info")
    @click.option("--overwrite", is_flag=True, default=False)
    def add_elem_info_cmd(overwrite: bool) -> None:
        click.echo(f"overwrite={overwrite}")

    runner = CliRunner()
    result_value_false = runner.invoke(cli, ["add-elem-info", "--overwrite", "False"])
    assert result_value_false.exit_code == 0
    assert "overwrite=False" in result_value_false.output

    result_no = runner.invoke(cli, ["add-elem-info", "--no-overwrite"])
    assert result_no.exit_code == 0
    assert "overwrite=False" in result_no.output


def test_toggle_with_non_no_negative_alias_accepts_no_prefix_and_values() -> None:
    cli = _make_group(normalize_bool_argv=_normalize_bool_argv_impl)

    @cli.command(name="define-layer")
    @click.option("--one-based/--zero-based", default=True)
    def define_layer_cmd(one_based: bool) -> None:
        click.echo(f"one_based={one_based}")

    runner = CliRunner()
    result_false = runner.invoke(cli, ["define-layer", "--one-based", "False"])
    assert result_false.exit_code == 0
    assert "one_based=False" in result_false.output

    result_no = runner.invoke(cli, ["define-layer", "--no-one-based"])
    assert result_no.exit_code == 0
    assert "one_based=False" in result_no.output


def test_parser_wrapper_bool_provider_enables_value_style_normalization() -> None:
    cli = _make_group(
        normalize_bool_argv=_normalize_bool_argv_impl,
        parser_wrapper_subcommands=frozenset({"extract"}),
        parser_wrapper_bool_option_providers={"extract": lambda: frozenset({"--verbose"})},
    )

    @cli.command(
        name="extract",
        context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
    )
    @click.pass_context
    def extract_cmd(ctx: click.Context) -> None:
        click.echo("|".join(ctx.args))

    runner = CliRunner()
    result = runner.invoke(cli, ["extract", "--verbose", "False"])
    assert result.exit_code == 0
    assert "--no-verbose" in result.output
    assert "--verbose|False" not in result.output



# [1] https://aip.scitation.org/doi/10.1063/1.471864
#     The generation and use of delocalized internal coordinates in geometry
#     optimization.
#     Baker, 1996

import numpy as np
import torch

from pysisyphus.intcoords import RedundantCoords
from pysisyphus.linalg import gram_schmidt


class DLC(RedundantCoords):
    def __init__(self, *args, full_set=True, **kwargs):
        super().__init__(*args, **kwargs)

        self.full_set = full_set
        self.set_active_set()

    @property
    def U(self):
        return self._U

    @U.setter
    def U(self, U):
        self._U = U

    @property
    def constraints(self):
        return self._constraints

    @constraints.setter
    def constraints(self, constraints):
        self.U = self.U_unconstrained.copy()
        constraints = np.array(constraints)
        constraints.flags.writeable = False
        # Constraint columns should be same length as columns of U.
        assert constraints.shape[0] == self.U.shape[0]
        self._constraints = constraints
        U_constrained = self.get_constrained_U(self._constraints)

        # # Replace the constraint-columns with zeros vectors, so the total
        # # number of coords doesn't change. This also zeros any force components
        # # that belong to the constraints.
        # zero_arr = np.zeros_like(constraints)
        # self.U = np.concatenate((zero_arr, U_constrained), axis=1)
        U_constrained = np.concatenate((constraints, U_constrained), axis=1)
        self.U = U_constrained

    def reset_constraints(self):
        self._constraints = tuple()
        self.U = self.U_unconstrained

    @property
    def coords(self):
        return self.U.T.dot(self.prim_coords)

    @property
    def B(self):
        """Wilson B-Matrix in the non-redundant subspace."""
        return self.U.T.dot(self.B_prim)

    def project_hessian(self, H):
        """As we already work in the non-redundant subspace we don't have
        to project/shift the hessian as we do it for simple redundant
        internal coordinates."""
        return H

    def transform_hessian(self, cart_hessian, int_gradient=None):
        """Transform Cartesian Hessian to DLC."""
        # Transform the DLC gradient to primitive coordinates
        if isinstance(cart_hessian, torch.Tensor):
            U = self._as_torch_like(self.U, cart_hessian)
            if int_gradient is not None:
                int_gradient = self._as_torch_like(int_gradient, cart_hessian)
                prim_gradient = (U * int_gradient).sum(dim=1)
            else:
                prim_gradient = None
            H = super().transform_hessian(cart_hessian, prim_gradient)
            return U.T @ H @ U
        if int_gradient is not None:
            prim_gradient = (self.U * int_gradient).sum(axis=1)
        else:
            prim_gradient = None
        H = super().transform_hessian(cart_hessian, prim_gradient)
        return self.U.T.dot(H).dot(self.U)

    def backtransform_hessian(self, *args, **kwargs):
        raise Exception("Check if we can just use the parents method.")

    def transform_int_step(self, step, *args, **kwargs):
        """As the transformation is done in primitive internal coordinates
        we convert the DLC back to primitive coordinates."""
        prim_step = (step * self.U).sum(axis=1)
        return super().transform_int_step(prim_step, *args, **kwargs)

    def get_active_set(self, B, inv_thresh=None):
        """See [5] between Eq. (7) and Eq. (8) for advice regarding
        the threshold."""
        if self.weighted:
            weights = np.array(
                [prim.weight(self.atoms, self.coords3d) for prim in self.primitives]
            )
            self.log(
                "Weighting B-matrix:\n"
                f"\tWeights: {np.array2string(weights, precision=4)}\n"
                f"\tmax(weights)={weights.max():.4f}, "
                f"min(weights)={weights.min():.4f}, ({len(weights)} primitives)"
            )
            B = np.diag(weights).dot(B)

        G = B.dot(B.T)
        eigvals, eigvectors = np.linalg.eigh(G)

        if inv_thresh is None:
            # The singular values of G=B B^T correspond to the square roots of the
            # eigenvalues of G.
            #
            # S = sqrt(w)
            # w = S**2
            #
            # To stay consistent with the SVD, we derive the eigenvalue threshold from
            # the SVD threshold.
            inv_thresh = self.svd_inv_thresh ** 2

        if self.full_set:
            use_inds = np.full_like(eigvals, False, dtype=bool)
            dof = 3 * len(self.atoms) - 6
            use_inds[-dof:] = True
        else:
            use_inds = np.abs(eigvals) > inv_thresh
        use_eigvals = eigvals[use_inds]
        min_eigval = use_eigvals.min()
        self.log(
            f"Diagonalizing G yielded {use_inds.sum()} DLCs. Smallest eigenvalue "
            f"is {min_eigval:.4e}"
        )
        return eigvectors[:, use_inds]

    def set_active_set(self):
        self.U = self.get_active_set(self.B_prim)
        # Keep a copy of the original active set, in case self.U gets
        # modified by constraint application.
        self.U_unconstrained = self.U.copy()
        self.original_U_shape = self.U_unconstrained.shape
        self._constraints = tuple()

    def project_primitive_on_active_set(self, prim_ind):
        prim_vec = np.zeros(self.U.shape[0])
        prim_vec[prim_ind] = 1
        c_proj = (np.einsum("i,ij->j", prim_vec, self.U) * self.U).sum(axis=1)
        c_proj /= np.linalg.norm(c_proj)
        return c_proj

    def get_constrained_U(self, constraint_vecs):
        # Constraints are organized in columns
        constr_num = constraint_vecs.shape[1]
        V = np.concatenate((constraint_vecs, self.U), axis=1)
        orthonormalized = gram_schmidt(V.T).T
        # During Gram-Schmidt a number of columns of U should have
        # dropped out. They are replaced by the constraint_vecs, so
        # the total shape of should not change.
        assert orthonormalized.shape[1] == self.original_U_shape[1]
        # Remove constraint vectors
        # [1] states that we somehow have to keep the constraint vectors
        # (or the corresponding unit vectors) for the iterative
        # back-transformation. Right now I don't understand why we would
        # have to do this ([1], p. 10 (200), right column).
        U_proj = orthonormalized[:, constr_num:]
        return U_proj

    def freeze_primitives(self, typed_prims):
        """Freeze primitive internal coordinates.

        Parameters
        ----------
        typed_prims : iterable of typed primitives
            Iterable containing typed_primitives, starting with a PrimType and
            followed by atom indices.
        """
        prim_indices = [self.get_index_of_typed_prim(tp) for tp in typed_prims]
        not_defined = [
            tp for tp, prim_ind in zip(typed_prims, prim_indices) if prim_ind is None
        ]
        assert (
            None not in prim_indices
        ), f"Some primitive internals are not defined ({not_defined})!"
        projected_primitives = np.array(
            [self.project_primitive_on_active_set(pi) for pi in prim_indices]
        ).T
        self.constraints = projected_primitives


class HDLC(DLC):
    def __init__(self, *args, **kwargs):
        kwargs["hybrid"] = True
        super().__init__(*args, **kwargs)



# [1] https://doi.org/10.1002/jcc.26495
#     Habershon, 2021

"""
prp a901cdfacc579eb63b193cbc9043212e8b57746f
pysis 340ab6105ac4156f0613b4d0e8f080d9f195530c
do_trans accidentally disabled in transtorque
"""

from functools import reduce
import itertools as it

import numpy as np

from pysisyphus.calculators import (
    HardSphere,
    TransTorque,
    AtomAtomTransTorque,
    Composite,
)
from pysisyphus.constants import BOHR2ANG
from pysisyphus.Geometry import Geometry
from pysisyphus.helpers import align_coords
from pysisyphus.helpers_pure import highlight_text
from pysisyphus.init_logging import init_logging
from pysisyphus.intcoords.setup import get_fragments, get_bond_sets
from pysisyphus.xyzloader import coords_to_trj, make_xyz_str


init_logging()


class SteepestDescent:
    def __init__(
        self,
        geom,
        max_cycles=1000,
        max_step=0.05,
        rms_force=0.05,
        rms_force_only=True,
        prefix=None,
        dump=False,
        dump_every=100,
        print_every=100,
    ):
        self.geom = geom
        self.max_cycles = max_cycles
        self.max_step = max_step
        self.rms_force = rms_force
        self.rms_force_only = rms_force_only
        if prefix is not None:
            prefix = f"{prefix}_"
        self.prefix = prefix
        self.dump = dump
        self.dump_every = dump_every
        self.print_every = print_every

        self.all_coords = np.zeros((max_cycles, self.geom.coords.size))

    def run(self):
        coords = self.geom.coords.copy()

        to_dump = []

        for i in range(self.max_cycles):
            self.all_coords[i] = coords.copy()
            if self.dump and (i % self.dump_every) == 0:
                to_dump.append(self.geom.as_xyz(cart_coords=coords))
            results = self.geom.get_energy_and_forces_at(coords)
            forces = results["forces"]
            norm = np.linalg.norm(forces)
            rms = np.sqrt(np.mean(forces**2))

            if i > 0:
                beta = forces.dot(forces) / self.prev_forces.dot(self.prev_forces)
                step = forces + beta * self.prev_step
            else:
                step = forces.copy()

            step *= min(self.max_step / np.abs(step).max(), 1)
            converged = rms <= self.rms_force

            # Append step information when not yet converged
            if i % self.print_every == 0 or converged:
                print(
                    f"{i:05d}: |forces|={norm: >12.6f} "
                    f"rms(forces)={np.sqrt(np.mean(forces**2)): >12.6f} "
                    f" |step|={np.linalg.norm(step): >12.6f}"
                )

            if converged:
                print(f"Converged in cycle {i}. Breaking.")
                break

            coords += step

            self.prev_step = step
            self.prev_forces = forces
        self.geom.coords = coords
        self.all_coords = self.all_coords[: i + 1]

        if to_dump:
            with open(f"{self.prefix}optimization_trj.xyz", "w") as handle:
                handle.write("\n".join(to_dump))


def get_fragments_and_bonds(geoms):
    if isinstance(geoms, Geometry) or len(geoms) == 1:
        geom = geoms
        atoms = geom.atoms
        coords3d = geom.coords3d
        bonds = [frozenset(bond) for bond in get_bond_sets(atoms, coords3d)]
        fragments = get_fragments(atoms, coords3d.flatten(), bond_inds=bonds)
        frag_inds = list(it.chain(*fragments))
        if len(frag_inds) != len(atoms):
            all_inds = list(range(len(atoms)))
            missing_inds = set(all_inds) - set(frag_inds)
            for mi in missing_inds:
                fragments.append(frozenset((mi,)))

        frag_bonds = [
            list(filter(lambda bond: bond <= frag, bonds)) for frag in fragments
        ]
        # frag_atoms = [[a for i, a in enumerate(atoms) if i in frag] for frag in fragments]

        # Assert that we do not have any interfragment bonds
        assert reduce((lambda x, y: x + len(y)), frag_bonds, 0) == len(bonds)
        union_geom = geom.copy(coord_type="cart")
    else:
        # Form union, determine consistent new indices for all atoms and calculate bonds
        raise Exception()

    # return fragments, frag_bonds, set(bonds), frag_atoms
    return fragments, frag_bonds, set(bonds), union_geom


def get_rot_mat(coords3d_1, coords3d_2, center=False):
    coords3d_1 = coords3d_1.copy().reshape(-1, 3)
    coords3d_2 = coords3d_2.copy().reshape(-1, 3)

    def _center(coords3d):
        return coords3d - coords3d.mean(axis=0)

    if center:
        coords3d_1 = _center(coords3d_1)
        coords3d_2 = _center(coords3d_2)

    tmp_mat = coords3d_1.T.dot(coords3d_2)
    U, W, Vt = np.linalg.svd(tmp_mat)
    rot_mat = U.dot(Vt)
    # Avoid reflections
    if np.linalg.det(rot_mat) < 0:
        U[:, -1] *= -1
        rot_mat = U.dot(Vt)
    return rot_mat


def get_steps_to_active_atom_mean(
    frag_lists, iter_frag_lists, ind_dict, coords3d, skip=True
):
    frag_num = len(frag_lists)
    steps = np.zeros((frag_num, 3))
    for m, frag_m in enumerate(frag_lists):
        step_m = np.zeros(3)
        for n, _ in enumerate(iter_frag_lists):
            if skip and m == n:
                continue
            active_inds = ind_dict[(n, m)]
            if len(active_inds) == 0:
                continue
            step_m += coords3d[active_inds].mean(axis=0)
        step_m /= frag_num
        steps[m] = step_m
    return steps


def report_frags(rgeom, pgeom, rfrags, pfrags, rbond_diff, pbond_diff):
    for name, geom in (("Reactant(s)", rgeom), ("Product(s)", pgeom)):
        print(f"{name}: {geom}\n\n{geom.as_xyz()}\n")

    def get_frag_atoms(geom, frag):
        atoms = geom.atoms
        return [atoms[i] for i in frag]

    for name, geom, frags in (("reactant", rgeom, rfrags), ("product", pgeom, pfrags)):
        print(f"{len(frags)} Fragment(s) in {name} image:\n")
        for frag in frags:
            frag_atoms = get_frag_atoms(geom, frag)
            frag_coords = geom.coords3d[list(frag)]
            frag_xyz = make_xyz_str(frag_atoms, frag_coords * BOHR2ANG)
            print(frag_xyz + "\n")

    def print_bonds(geom, bonds):
        for from_, to_ in bonds:
            from_atom, to_atom = [geom.atoms[i] for i in (from_, to_)]
            print(f"\t({from_: >3d}{from_atom} - {to_: >3d}{to_atom})")

    print("Bonds broken in reactant image:")
    print_bonds(rgeom, rbond_diff)
    print()
    print("Bonds formed in product image:")
    print_bonds(pgeom, pbond_diff)
    print()


def report_mats(name, mats):
    for (m, n), indices in mats.items():
        print(f"{name}({m}, {n}): {indices}")
    print()


def center_fragments(frag_list, geom):
    c3d = geom.coords3d
    for frag in frag_list:
        mean = c3d[frag].mean(axis=0)
        c3d[frag] -= mean[None, :]


def get_which_frag(frags):
    which_frag = dict()
    for frag_ind, frag in enumerate(frags):
        which_frag.update({atom_ind: frag_ind for atom_ind in frag})
    return which_frag


def form_A(frags, which_frag, formed_bonds):
    """Construct the A-matrices.

    AR[(m, n)] (AP[(m, n)]) contains the subset of atoms in Rm (Pm) that forms
    bonds with Rn (Pn).
    """
    A = dict()
    for m, n in formed_bonds:
        key = (which_frag[m], which_frag[n])
        A.setdefault(key, list()).append(m)
        A.setdefault(key[::-1], list()).append(n)
    return A


def rotate_inplace(frags, union, bonds):
    # which_frag could also be calculated outside of this function
    which_frag = dict()
    for i, frag in enumerate(frags):
        for ind in frag:
            which_frag[ind] = i
    AR = form_A(frags, which_frag, bonds)

    def form_G(A):
        G = dict()
        for (m, n), inds in A.items():
            G.setdefault(m, set())
            G[m] |= set(inds)

        for k, v in G.items():
            G[k] = list(v)
            assert len(v) > 0
        return G

    GR = form_G(AR)

    # Rotate R fragments
    alphas = get_steps_to_active_atom_mean(frags, frags, AR, union.coords3d)
    gammas = np.zeros_like(alphas)
    for m, rfrag in enumerate(frags):
        Gm = GR[m]
        gammas[m] = union.coords3d[Gm].mean(axis=0)
    r_means = np.array([union.coords3d[frag].mean(axis=0) for frag in frags])

    for m, rfrag in enumerate(frags):
        gm = r_means[m]
        rot_mat = get_rot_mat(gammas[m] - gm, alphas[m] - gm)
        rot_coords = (union.coords3d[rfrag] - gm).dot(rot_mat)
        union.coords3d[rfrag] = rot_coords + gm - rot_coords.mean(axis=0)


CONFIG = {
    "s2_hs_kappa": 1.0,
    "s4_hs_kappa": 50.0,
    "s4_v_kappa": 1.0,
    "s4_w_kappa": 1.0,
    "s5_v_kappa": 1.0,
    "s5_w_kappa": 3.0,
    "s5_hs_kappa": 10.0,
    "s5_z_kappa": 2.0,
    "s5_trans": True,
    "s5_rms_force": 0.01,
}


def precon_pos_rot(reactants, products, prefix=None, config=CONFIG):
    c = config

    if prefix is None:
        prefix = ""

    def make_fn(fn):
        return prefix + fn

    rfrags, rfrag_bonds, rbonds, runion = get_fragments_and_bonds(reactants)
    pfrags, pfrag_bonds, pbonds, punion = get_fragments_and_bonds(products)

    pbond_diff = pbonds - rbonds  # Present in product(s)
    rbond_diff = rbonds - pbonds  # Present in reactant(s)
    involved_atoms = set(tuple(it.chain(*pbond_diff)))
    involved_atoms |= set(tuple(it.chain(*rbond_diff)))

    which_rfrag = get_which_frag(rfrags)
    which_pfrag = get_which_frag(pfrags)

    rfrag_lists = [list(frag) for frag in rfrags]
    pfrag_lists = [list(frag) for frag in pfrags]

    report_frags(runion, punion, rfrags, pfrags, rbond_diff, pbond_diff)

    def form_C(m_frags, n_frags):
        """Construct the C-matrices.

        Returns a dict with (m, n) keys, containing the respective
        unions of rectant fragment n and product fragment m.
        """
        C = dict()
        for m, m_frag in enumerate(m_frags):
            for n, n_frag in enumerate(n_frags):
                C[(m, n)] = list(m_frag & n_frag)
        return C

    CR = form_C(rfrags, pfrags)
    assert len(set(it.chain(*CR.values()))) == len(runion.atoms)
    CP = {(n, m): union for (m, n), union in CR.items()}
    print("CR(m, n), subset of atoms in molecule Rn which are in Pm after reaction.")
    report_mats("CR", CR)
    print("CP(m, n), subset of atoms in molecule Pn which are in Rm before reaction.")
    report_mats("CP", CP)

    def form_B(C):
        """Construct the B-matrices.

        Returns a dict with (m, n) keys, containing the respective
        subsets of C[(m, n)] that acutally participate in bond-breaking/forming.
        """
        B = dict()
        for (m, n), union in C.items():
            key = (m, n)
            B.setdefault(key, set())
            B[key] |= set(union) & involved_atoms
        for k, v in B.items():
            B[k] = list(v)
        return B

    BR = form_B(CR)
    BP = form_B(CP)
    print(
        "BR(m, n), subset of atoms in CRnm actually involved in bond forming/breaking."
    )
    report_mats("BR", BR)
    print(
        "BP(m, n), subset of atoms in CPnm actually involved in bond forming/breaking."
    )
    report_mats("BP", BP)

    AR = form_A(rfrags, which_rfrag, pbond_diff)
    AP = form_A(pfrags, which_pfrag, rbond_diff)
    print("AR(m, n), subset of atoms in Rm that form bonds to atoms in Rn.")
    report_mats("AR", AR)
    print(
        "AP(m, n), subset of atoms in Pm which had bonds with Pn (formerly bonded in R)."
    )
    report_mats("AP", AP)

    def form_G(A):
        G = dict()
        for (m, n), inds in A.items():
            G.setdefault(m, set())
            G[m] |= set(inds)

        for k, v in G.items():
            G[k] = list(v)
            assert len(v) > 0
        return G

    GR = form_G(AR)
    # GP = form_G(AP)
    print(f"GR: {GR}")
    # print(f"GP: {GP}")

    # Initial, centered, coordinates and 5 stages
    r_coords = np.zeros((6, runion.coords.size))
    p_coords = np.zeros((6, punion.coords.size))

    def backup_coords(stage):
        assert 0 <= stage < 6
        r_coords[stage] = runion.coords.copy()
        p_coords[stage] = punion.coords.copy()

    """
    STAGE 1
    Initial positioning of reactant and product molecules
    """

    # Center fragments at their geometric average
    center_fragments(rfrag_lists, runion)
    center_fragments(pfrag_lists, punion)

    backup_coords(0)

    # Translate reactant molecules
    alphas = get_steps_to_active_atom_mean(
        rfrag_lists, rfrag_lists, AR, runion.coords3d
    )
    for rfrag, alpha in zip(rfrag_lists, alphas):
        runion.coords3d[rfrag] += alpha

    # Translate product molecules
    betas = get_steps_to_active_atom_mean(
        pfrag_lists, rfrag_lists, BR, punion.coords3d, skip=False
    )
    sigmas = get_steps_to_active_atom_mean(
        pfrag_lists, rfrag_lists, CR, punion.coords3d, skip=False
    )
    bs_half = (betas + sigmas) / 2
    for pfrag, bsh in zip(pfrag_lists, bs_half):
        punion.coords3d[pfrag] += bsh

    backup_coords(1)
    print()

    """
    STAGE 2
    Intra-image Inter-molecular Hard-Sphere forces
    """

    print(highlight_text("Stage 2, Hard-Sphere Forces"))

    s2_hs_kappa = c["s2_hs_kappa"]

    def hardsphere_sd_opt(geom, frag_lists, title):
        print(highlight_text(title, level=1))
        calc = HardSphere(geom, frag_lists, kappa=s2_hs_kappa)
        geom.set_calculator(calc)
        opt_kwargs = {
            "max_cycles": 1000,
            "max_step": 0.5,
            "rms_force": 0.05,
        }
        opt = SteepestDescent(geom, **opt_kwargs)
        opt.run()

    hardsphere_sd_opt(runion, rfrag_lists, "Reactants")
    hardsphere_sd_opt(punion, pfrag_lists, "Products")

    backup_coords(2)
    print()

    """
    STAGE 3
    Initial orientation of molecules
    """

    print(highlight_text("Stage 3, Initial Orientation"))

    # Rotate R fragments
    # TODO: refactor to use rotate_inplace()
    if len(rfrag_lists) > 1:
        alphas = get_steps_to_active_atom_mean(
            rfrag_lists, rfrag_lists, AR, runion.coords3d
        )
        gammas = np.zeros_like(alphas)
        for m, rfrag in enumerate(rfrag_lists):
            Gm = GR[m]
            gammas[m] = runion.coords3d[Gm].mean(axis=0)
        r_means = np.array([runion.coords3d[frag].mean(axis=0) for frag in rfrag_lists])

        for m, rfrag in enumerate(rfrag_lists):
            gm = r_means[m]
            rot_mat = get_rot_mat(gammas[m] - gm, alphas[m] - gm)
            rot_coords = (runion.coords3d[rfrag] - gm).dot(rot_mat)
            runion.coords3d[rfrag] = rot_coords + gm - rot_coords.mean(axis=0)

    Ns = [0] * len(pfrag_lists)
    for (m, n), CPmn in CP.items():
        Ns[m] += len(CPmn)

    # Rotate P fragments
    for m, pfrag in enumerate(pfrag_lists):
        pc3d = punion.coords3d[pfrag]
        gm = pc3d.mean(axis=0)
        r0Pm = pc3d - gm[None, :]
        mu_Pm = np.zeros_like(r0Pm)
        N = Ns[m]
        for n, rfrag in enumerate(rfrag_lists):
            # Skip rotation of 1-atom fragments
            if len(rfrag) == 1:
                continue
            CPmn = CP[(m, n)]
            RPmRn = get_rot_mat(
                punion.coords3d[CPmn], runion.coords3d[CPmn], center=True
            )
            print(f"m={m}, n={n}, len(CPmn)={len(CPmn)}")
            # Eq. (A2) in [1]
            r0Pmn = np.einsum("ij,jk->ki", RPmRn, r0Pm.T)
            mu_Pm += len(CPmn) ** 2 / N * r0Pmn
        rot_mat = get_rot_mat(r0Pm, mu_Pm, center=True)
        rot_coords = r0Pm.dot(rot_mat)
        punion.coords3d[pfrag] = rot_coords + gm - rot_coords.mean(axis=0)

    backup_coords(3)
    print()

    """
    STAGE 4
    Alignment of reactive atoms

    This stage involves three forces: hard-sphere forces and two kinds
    of average translational (^t) and rotational (^r) forces (v and w,
    (A3) - (A5) in [1]).

    v^t and v^r arise from atoms in A^Rnm and A^Rmn, that is atoms that
    participate in bond forming/breaking in R. The translational force
    is usually attractive, which is counteracted by the repulsive hard-sphere
    forces.
    """

    print(highlight_text("Stage 4, Alignment Of Reactive Atoms"))

    def composite_sd_opt(geom, keys_calcs, title, rms_force=0.05):
        print(highlight_text(title, level=1))
        final = " + ".join([k for k in keys_calcs.keys()])
        calc = Composite(final, keys_calcs=keys_calcs)
        geom.set_calculator(calc)
        opt_kwargs = {
            "max_step": 0.05,
            "max_cycles": 2000,
            "rms_force": rms_force,
        }
        opt = SteepestDescent(geom, **opt_kwargs)
        opt.run()

    def get_vr_trans_torque(kappa=1.0, do_trans=True):
        return TransTorque(
            rfrag_lists, rfrag_lists, AR, AR, kappa=kappa, do_trans=do_trans
        )

    def r_weight_func(m, n, a, b):
        """As required for (A5) in [1]."""
        return 1 if a in BR[(m, n)] else 0.5

    def get_wr_trans_torque(kappa=1.0, do_trans=True):
        return TransTorque(
            rfrag_lists,
            pfrag_lists,
            CR,
            CP,
            weight_func=r_weight_func,
            skip=False,
            b_coords3d=punion.coords3d,
            kappa=kappa,
            do_trans=do_trans,
        )

    def get_vp_trans_torque(kappa=1.0, do_trans=True):
        return TransTorque(
            pfrag_lists, pfrag_lists, AP, AP, kappa=kappa, do_trans=do_trans
        )

    def p_weight_func(m, n, a, b):
        """As required for (A5) in [1]."""
        return 1 if a in BP[(m, n)] else 0.5

    def get_wp_trans_torque(kappa=1.0, do_trans=True):
        return TransTorque(
            pfrag_lists,
            rfrag_lists,
            CP,
            CR,
            weight_func=p_weight_func,
            skip=False,
            b_coords3d=runion.coords3d,
            kappa=kappa,
            do_trans=do_trans,
        )

    s4_hs_kappa = c["s4_hs_kappa"]
    s4_v_kappa = c["s4_v_kappa"]
    s4_w_kappa = c["s4_w_kappa"]

    vr_trans_torque = get_vr_trans_torque(kappa=s4_v_kappa)
    wr_trans_torque = get_wr_trans_torque(kappa=s4_w_kappa)
    r_keys_calcs = {
        "hardsphere": HardSphere(runion, rfrag_lists, kappa=s4_hs_kappa),
        "v": vr_trans_torque,
        "w": wr_trans_torque,
    }
    composite_sd_opt(runion, r_keys_calcs, "Reactants")

    vp_trans_torque = get_vp_trans_torque(kappa=s4_v_kappa)
    wp_trans_torque = get_wp_trans_torque(kappa=s4_w_kappa)
    p_keys_calcs = {
        "hardsphere": HardSphere(punion, pfrag_lists, kappa=s4_hs_kappa),
        "v": vp_trans_torque,
        "w": wp_trans_torque,
    }
    composite_sd_opt(punion, p_keys_calcs, "Products")

    backup_coords(4)
    print()

    """
    STAGE 5
    Refinement of atomic positions using further hard-sphere forces.
    """

    print(highlight_text("Stage 5, Refinement"))

    s5_v_kappa = c["s5_v_kappa"]
    s5_w_kappa = c["s5_w_kappa"]
    s5_hs_kappa = c["s5_hs_kappa"]
    s5_z_kappa = c["s5_z_kappa"]
    s5_trans = c["s5_trans"]
    s5_rms_force = c["s5_rms_force"]

    vr_trans_torque = get_vr_trans_torque(kappa=s5_v_kappa, do_trans=s5_trans)
    wr_trans_torque = get_wr_trans_torque(kappa=s5_w_kappa, do_trans=s5_trans)
    zr_aa_trans_torque = AtomAtomTransTorque(runion, rfrag_lists, AR, kappa=s5_z_kappa)
    r_keys_calcs = {
        "v": vr_trans_torque,
        "w": wr_trans_torque,
        "hardsphere": HardSphere(runion, rfrag_lists, kappa=s5_hs_kappa),
        "z": zr_aa_trans_torque,
    }
    composite_sd_opt(runion, r_keys_calcs, "Reactants", rms_force=s5_rms_force)

    vp_trans_torque = get_vp_trans_torque(kappa=s5_v_kappa, do_trans=s5_trans)
    wp_trans_torque = get_wp_trans_torque(kappa=s5_w_kappa, do_trans=s5_trans)
    zp_aa_trans_torque = AtomAtomTransTorque(punion, pfrag_lists, AP, kappa=s5_z_kappa)
    p_keys_calcs = {
        "v": vp_trans_torque,
        "w": wp_trans_torque,
        "hardsphere": HardSphere(punion, pfrag_lists, kappa=s5_hs_kappa),
        "z": zp_aa_trans_torque,
    }
    composite_sd_opt(punion, p_keys_calcs, "Products", rms_force=s5_rms_force)

    backup_coords(5)
    print()

    with open(make_fn("s5_trj.xyz"), "w") as handle:
        handle.write("\n".join([geom.as_xyz() for geom in (runion, punion)]))

    def dump_stages(fn, atoms, coords_list):
        align_coords(coords_list)
        comments = [f"Stage {i}" for i in range(coords_list.shape[0])]
        fn = make_fn(fn)
        coords_to_trj(fn, atoms, coords_list, comments=comments)

    dump_stages("r_coords_trj.xyz", runion.atoms, r_coords)
    dump_stages("p_coords_trj.xyz", punion.atoms, p_coords)

    runion.set_calculator(None)
    punion.set_calculator(None)
    return runion, punion


def run_precontr(reactant_geom, product_geom, **kwargs):
    print(
        highlight_text("Preconditioning of Translation & Rotation")
        + "\n\nPlease cite https://doi.org/10.1002/jcc.26495\n"
    )

    return precon_pos_rot(reactant_geom, product_geom, **kwargs)
